from __future__ import annotations

from ...core import AbstractToolbarSettings


__all__ = [
    'MiscSettings'
]


class MiscSettings(AbstractToolbarSettings):
    __slots__ = ()

    _add_to_tab = False

    SAVE_TEMPLATE = '{script_name}_{frame}'

    def setup_ui(self) -> None:
        super().setup_ui()

    def set_defaults(self) -> None:
        pass



from __future__ import annotations

import logging
import os
import signal
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Sequence

from PyQt6.QtWidgets import QApplication

from .core.logger import set_log_level, setup_logger
# import vsenv as early as possible:
# This is so other modules cannot accidentally use and lock us into a different policy.
from .core.vsenv import set_vsengine_loop
from .main import MainWindow
from .plugins.install import install_plugins, plugins_commands, uninstall_plugins

__all__ = [
    'main'
]


def main(_args: Sequence[str] | None = None, no_exit: bool = False) -> None:
    from .utils import exit_func

    parser = ArgumentParser(prog='VSPreview')
    parser.add_argument(
        'script_path_or_command', type=str, nargs='?',
        help=f'Path to Vapoursynth script or plugins command {",".join(plugins_commands)}'
    )
    parser.add_argument(
        'plugins', type=str, nargs='*',
        help='Plugins to install/uninstall/update'
    )
    parser.add_argument(
        '--version', '-v', action='version', version='%(prog)s 0.2b'
    )
    parser.add_argument(
        '--preserve-cwd', '-c', action='store_true', help='do not chdir to script parent directory'
    )
    parser.add_argument(
        '--arg', '-a', type=str, action='append', metavar='key=value', help='Argument to pass to the script environment'
    )
    parser.add_argument('-f', '--frame', type=int, help='Frame to load initially (defaults to 0)')
    parser.add_argument(
        '--vscode-setup', type=str, choices=['override', 'append', 'ignore'], nargs='?', const='append',
        help='Installs launch settings in cwd\'s .vscode'
    )
    parser.add_argument(
        "--verbose", help="Set the logging to verbose.", action="store_true"
    )
    parser.add_argument(
        "--force", help="Force the install of a plugin even if it exists already.", action="store_true"
    )
    parser.add_argument(
        "--no-deps", help="Ignore downloading dependencies.", action="store_true"
    )

    args = parser.parse_args(_args)

    setup_logger()

    if args.verbose:
        from vstools import VSDebug
        set_log_level(logging.DEBUG, logging.DEBUG)
        VSDebug(use_logging=True)
    else:
        set_log_level(logging.WARNING)

    if args.vscode_setup is not None:
        from .api.other import install_vscode_launch

        install_vscode_launch(args.vscode_setup)

        return exit_func(0, no_exit)

    script_path_or_command = args.script_path_or_command

    if not script_path_or_command and not (args.plugins and (script_path_or_command := next(iter(args.plugins)))):
        logging.error('Script path required.')
        return exit_func(1, no_exit)

    if (command := script_path_or_command) in plugins_commands:
        if not args.plugins:
            logging.error('You must provide at least one plugin!')
            return exit_func(1, no_exit)

        set_log_level(logging.INFO)

        plugins = list(args.plugins)

        if command == 'install':
            install_plugins(plugins, args.force, args.no_deps)
        elif command == 'uninstall':
            uninstall_plugins(plugins)
        elif command == 'update':
            uninstall_plugins(plugins, True)
            install_plugins(plugins, True, args.no_deps)

        return exit_func(0, no_exit)

    script_path = Path(script_path_or_command).resolve()
    if not script_path.exists():
        logging.error('Script path is invalid.')
        return exit_func(1, no_exit)

    if not args.preserve_cwd:
        os.chdir(script_path.parent)

    first_run = not hasattr(main, 'app')

    if first_run:
        main.app = QApplication(sys.argv)
        set_vsengine_loop()
    else:
        from .core.vsenv import make_environment, get_current_environment
        make_environment()
        get_current_environment().use()

    signal.signal(signal.SIGINT, signal.SIG_DFL)

    main.main_window = MainWindow(Path(os.getcwd()) if args.preserve_cwd else script_path.parent, no_exit)
    main.main_window.load_script(
        script_path, [tuple(a.split('=', maxsplit=1)) for a in args.arg or []], False, args.frame or None
    )
    main.main_window.show()

    ret_code = main.app.exec()

    if no_exit:
        from .core.vsenv import _dispose

        main.main_window.hide()

        _dispose()

    return exit_func(ret_code, no_exit)


if __name__ == '__main__':
    main()



from __future__ import annotations

from collections import deque
from concurrent.futures import Future
from fractions import Fraction
from functools import partial
from math import floor
from time import perf_counter_ns
from typing import TYPE_CHECKING, Any, Mapping, cast

import vapoursynth as vs
from PyQt6.QtCore import QKeyCombination, Qt
from PyQt6.QtWidgets import QComboBox, QSlider

from ...core import (
    AbstractToolbar, AudioOutput, CheckBox, ComboBox, DoubleSpinBox, Frame, FrameEdit, PushButton, Time, TimeEdit,
    Timer, try_load
)
from ...core.types import video
from ...models import AudioOutputs
from ...utils import qt_silent_call
from .settings import PlaybackSettings

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'PlaybackToolbar'
]


def _del_future(f: Future[vs.VideoFrame]) -> None:
    f0 = f.result()
    del f, f0


class PlaybackToolbar(AbstractToolbar):
    storable_attrs = ('audio_muted', 'audio_outputs', 'volume')

    __slots__ = (
        *storable_attrs, 'play_timer', 'fps_timer', 'fps_history', 'current_fps',
        'seek_n_frames_b_button', 'seek_to_prev_button', 'play_pause_button',
        'seek_to_next_button', 'seek_n_frames_f_button', 'play_n_frames_button',
        'seek_frame_control', 'seek_time_control',
        'fps_spinbox', 'fps_unlimited_checkbox', 'fps_variable_checkbox', 'fps_reset_button',
        'play_start_time', 'play_start_frame', 'play_end_time',
        'play_end_frame', 'play_buffer', 'toggle_button', 'play_timer_audio',
        'current_audio_frame', 'play_buffer_audio', 'audio_outputs',
        'audio_outputs_combobox', 'seek_to_start_button', 'seek_to_end_button',
        'audio_volume_slider'
    )

    settings: PlaybackSettings

    audio_outputs: AudioOutputs
    seek_frame_control: FrameEdit
    fps_spinbox: DoubleSpinBox

    def __init__(self, main: MainWindow) -> None:
        super().__init__(main, PlaybackSettings(self))
        self.setup_ui()

        self.play_buffer = deque[tuple[int, Future[vs.VideoFrame]]]()
        self.play_timer = Timer(timeout=self._show_next_frame, timerType=Qt.TimerType.PreciseTimer)

        self.play_timer_audio = Timer(timeout=self._play_next_audio_frame, timerType=Qt.TimerType.PreciseTimer)

        self.current_audio_output = cast(AudioOutput | None, None)
        self.current_audio_frame = Frame(0)
        self.play_buffer_audio = deque[Future[vs.AudioFrame]]()

        self.fps_history = deque[int]([], int(self.settings.FPS_AVERAGING_WINDOW_SIZE) + 1)
        self.current_fps = 0.0
        self.fps_timer = Timer(
            timeout=lambda: self.fps_spinbox.setValue(float(self.current_fps)), timerType=Qt.TimerType.PreciseTimer
        )

        self.play_start_time: int | None = None
        self.play_start_frame = Frame(0)
        self.play_end_time = 0
        self.play_end_frame = Frame(0)
        self.audio_outputs = cast(AudioOutputs, [])
        self.last_frame = Frame(0)

        self.setVolume(50, True)
        self.setMute(True)

        self.main.timeline.clicked.connect(self.on_timeline_clicked)

        self.add_shortcuts()

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.seek_to_start_button = PushButton(
            '⏮', self, tooltip='Seek to First Frame', clicked=self.seek_to_start
        )

        self.seek_n_frames_b_button = PushButton(
            '⏪', self, tooltip='Seek N Frames Backwards',
            clicked=lambda _: self.seek_offset(self.seek_frame_control.value() * -1)  # type: ignore
        )

        self.seek_to_prev_button = PushButton(
            '⏪', self, tooltip='Seek 1 Frame Backwards', clicked=lambda _: self.seek_offset(-1)
        )

        self.play_pause_button = PushButton(
            '⏯', self, tooltip='Play/Pause', checkable=True, clicked=self.on_play_pause_clicked
        )

        self.seek_to_next_button = PushButton(
            '⏩', self, tooltip='Seek 1 Frame Forward', clicked=lambda _: self.seek_offset(1)
        )

        self.seek_n_frames_f_button = PushButton(
            '⏩', self, tooltip='Seek N Frames Forward',
            clicked=lambda _: self.seek_offset(self.seek_frame_control.value())  # type: ignore
        )

        self.seek_to_end_button = PushButton(
            '⏭', self, tooltip='Seek to Last Frame', clicked=self.seek_to_end
        )

        self.seek_frame_control = FrameEdit(
            self, 1, value=self.settings.SEEK_STEP, tooltip='Seek N Frames Step',
            valueChanged=self.on_seek_frame_changed
        )

        self.play_n_frames_button = PushButton(
            '⏯', self, tooltip='Play N Frames', checkable=True, clicked=self.on_play_n_frames_clicked
        )

        self.seek_time_control = TimeEdit(self, valueChanged=self.on_seek_time_changed)

        self.fps_spinbox = DoubleSpinBox(self, valueChanged=self.on_fps_changed)
        self.fps_spinbox.setRange(0.001, 9999.0)
        self.fps_spinbox.setDecimals(3)
        self.fps_spinbox.setSuffix(' fps')

        self.fps_reset_button = PushButton('Reset FPS', self, clicked=self.reset_fps)

        self.fps_unlimited_checkbox = CheckBox('Unlimited FPS', self, stateChanged=self.on_fps_unlimited_changed)

        self.fps_variable_checkbox = CheckBox('Variable FPS', self, stateChanged=self.on_fps_variable_changed)

        self.mute_button = PushButton('', self, clicked=self.on_mute_clicked)
        self.mute_button.setFixedWidth(18)

        self.audio_outputs_combobox = ComboBox[AudioOutput](
            self, editable=True, insertPolicy=QComboBox.InsertPolicy.InsertAtCurrent,
            duplicatesEnabled=True, sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents
        )

        self.audio_volume_slider = QSlider(Qt.Orientation.Horizontal, valueChanged=self.setVolume)  # type: ignore
        self.audio_volume_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.audio_volume_slider.setFixedWidth(120)
        self.audio_volume_slider.setRange(0, 100)
        self.audio_volume_slider.setPageStep(5)

        self.hlayout.addWidgets([
            self.seek_to_start_button, self.seek_n_frames_b_button, self.seek_to_prev_button,
            self.play_pause_button,
            self.seek_to_next_button, self.seek_n_frames_f_button, self.seek_to_end_button,
            self.seek_frame_control, self.play_n_frames_button,
            self.seek_time_control,
            self.fps_spinbox, self.fps_reset_button,
            self.fps_unlimited_checkbox, self.fps_variable_checkbox,
            self.get_separator(),
            self.audio_outputs_combobox, self.mute_button, self.audio_volume_slider
        ])

        self.hlayout.addStretch()

    def add_shortcuts(self) -> None:
        self.main.add_shortcut(Qt.Key.Key_Space, self.play_pause_button.click)
        self.main.add_shortcut(Qt.Key.Key_Left, self.seek_to_prev_button.click)
        self.main.add_shortcut(Qt.Key.Key_Right, self.seek_to_next_button.click)
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.SHIFT, Qt.Key.Key_Left).toCombined(), self.seek_n_frames_b_button.click
        )
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.SHIFT, Qt.Key.Key_Right).toCombined(), self.seek_n_frames_f_button.click
        )
        self.main.add_shortcut(Qt.Key.Key_PageUp, self.seek_n_frames_b_button.click)
        self.main.add_shortcut(Qt.Key.Key_PageDown, self.seek_n_frames_f_button.click)
        self.main.add_shortcut(Qt.Key.Key_Home, self.seek_to_start_button.click)
        self.main.add_shortcut(Qt.Key.Key_End, self.seek_to_end_button.click)

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        qt_silent_call(self.seek_frame_control.setMaximum, self.main.current_output.total_frames)
        qt_silent_call(self.seek_time_control.setMaximum, self.main.current_output.total_time)
        qt_silent_call(self.seek_time_control.setMinimum, Time(Frame(1)))
        qt_silent_call(self.seek_time_control.setValue, Time(self.seek_frame_control.value()))
        qt_silent_call(self.fps_spinbox.setValue, float(self.main.current_output.play_fps))

    def rescan_outputs(self, outputs: AudioOutputs | None = None) -> None:
        self.audio_outputs = outputs if isinstance(outputs, AudioOutputs) else AudioOutputs(self.main)
        self.audio_outputs_combobox.setModel(self.audio_outputs)

    def get_true_fps(self, n: int | Frame, frameprops: vs.FrameProps, force: bool = False) -> Fraction:
        if (
            hasattr(self.main.current_output, 'got_timecodes') and self.main.current_output.got_timecodes and not force
        ):
            return Fraction(self.main.current_output.timecodes[int(n)])

        if any({x not in frameprops for x in {'_DurationDen', '_DurationNum'}}):
            raise RuntimeError(
                'Playback: DurationDen and DurationNum frame props are needed for VFR clips!'
            )
        return Fraction(frameprops['_DurationDen'], frameprops['_DurationNum'])  # type: ignore

    def allocate_buffer(self, is_alpha: bool = False) -> None:
        if is_alpha:
            play_buffer_size = int(min(
                self.settings.playback_buffer_size,
                int(self.main.current_output.total_frames - self.main.current_output.last_showed_frame - 1) * 2
            ))
            play_buffer_size -= play_buffer_size % 2
        else:
            play_buffer_size = int(min(
                self.settings.playback_buffer_size,
                int(self.main.current_output.total_frames - self.main.current_output.last_showed_frame - 1)
            ))

        self.play_buffer = deque([], play_buffer_size)

    @property
    def got_debug_fps(self) -> bool:
        return hasattr(self.main.toolbars, 'debug') and self.main.toolbars.debug.settings.DEBUG_PLAY_FPS

    def play(self, stop_at_frame: int | Frame | None = None) -> None:
        if (
            self.main.current_output.last_showed_frame > self.main.current_output.total_frames
        ):
            return

        if not video.PACKING_TYPE.can_playback:
            import logging
            logging.warn(
                f'The current backend ({video.PACKING_TYPE.name}) can\'t playback! Install akarin or libp2p plugins.'
            )
            return

        if self.main.statusbar.label.text() == 'Ready':
            self.main.statusbar.label.setText('Playing')

        if self.main.current_output.prepared.alpha is None:
            self.allocate_buffer(False)
            for i in range(cast(int, self.play_buffer.maxlen)):
                nextFrame = int(Frame(self.main.current_output.last_showed_frame) + i + 1)
                if nextFrame >= self.main.current_output.total_frames:
                    break
                self.play_buffer.appendleft(
                    (nextFrame, self.main.current_output.prepared.clip.get_frame_async(nextFrame))  # type: ignore
                )
        else:
            self.allocate_buffer(True)
            for i in range(cast(int, self.play_buffer.maxlen) // 2):
                nextFrame = int(Frame(self.main.current_output.last_showed_frame) + i + 1)
                if nextFrame >= self.main.current_output.total_frames:
                    break
                self.play_buffer.appendleft(
                    (nextFrame, self.main.current_output.prepared.clip.get_frame_async(nextFrame))  # type: ignore
                )
                self.play_buffer.appendleft(
                    (nextFrame, self.main.current_output.prepared.alpha.get_frame_async(nextFrame))  # type: ignore
                )

        self.last_frame = Frame(stop_at_frame or (self.main.current_output.total_frames - 1))

        if self.fps_unlimited_checkbox.isChecked() or self.got_debug_fps:
            self.mute_button.setChecked(True)
            self.play_timer.start(0)
            if self.got_debug_fps:
                self.play_start_time = perf_counter_ns()
                self.play_start_frame = Frame(self.main.current_output.last_showed_frame)
            else:
                self.fps_timer.start(self.settings.FPS_REFRESH_INTERVAL)
        else:
            if self.fps_variable_checkbox.isChecked() and self.main.current_output._stateset:
                fps = self.get_true_fps(self.last_frame, self.main.current_output.props)
            else:
                fps = self.main.current_output.play_fps

            self.play_timer.start(floor(1000 / fps))

        self.current_audio_output = self.audio_outputs_combobox.currentValue()

        if not self.audio_muted and self.current_audio_output is not None:
            self.play_audio()

    def play_audio(self) -> None:
        if not len(self.audio_outputs):
            return

        self.current_audio_output = self.audio_outputs_combobox.currentValue()

        if not self.current_audio_output.vs_output:
            return

        self.audio_outputs_combobox.setEnabled(False)
        self.current_audio_frame = self.current_audio_output.to_frame(
            Time(self.main.current_output.last_showed_frame)
        )

        self.current_audio_output.render_audio_frame(self.current_audio_frame)
        self.current_audio_output.render_audio_frame(self.current_audio_frame + Frame(1))
        self.current_audio_output.render_audio_frame(self.current_audio_frame + Frame(2))

        self.play_buffer_audio = deque([], int(min(
            self.settings.playback_buffer_size,
            int(self.current_audio_output.total_frames - self.current_audio_frame - 1)
        )))

        for i in range(2, cast(int, self.play_buffer_audio.maxlen)):
            self.play_buffer_audio.appendleft(
                self.current_audio_output.vs_output.get_frame_async(  # type: ignore
                    int(self.current_audio_frame + i + 1)
                )
            )

        self.play_timer_audio.start(
            floor(
                1000 / self.current_audio_output.fps / self.main.current_output.play_fps * self.main.current_output.fps
            )
        )

    def _show_next_frame(self) -> None:
        if not self.main.current_output.prepared:
            return

        if self.last_frame <= self.main.current_output.last_showed_frame:
            return self.stop()

        n_frames = 1 if self.main.current_output.prepared.alpha is None else 2
        next_buffered_frame = self.main.current_output.last_showed_frame + (
            self.settings.playback_buffer_size // n_frames
        )

        try:
            frames_futures = [
                (x[0], x[1].result()) for x in [
                    self.play_buffer.pop() for _ in range(n_frames)
                ]
            ]
        except IndexError:
            return self.play_pause_button.click()

        if next_buffered_frame < self.main.current_output.total_frames:
            self.play_buffer.appendleft((
                next_buffered_frame,
                self.main.current_output.prepared.clip.get_frame_async(next_buffered_frame)  # type: ignore
            ))

            if self.main.current_output.prepared.alpha is not None:
                self.play_buffer.appendleft((
                    next_buffered_frame,
                    self.main.current_output.prepared.alpha.get_frame_async(next_buffered_frame)  # type: ignore
                ))

        curr_frame = Frame(frames_futures[0][0])

        if self.fps_variable_checkbox.isChecked():
            self.current_fps = float(self.get_true_fps(curr_frame.value, frames_futures[0][1].props))
            self.play_timer.start(floor(1000 / self.current_fps))
            self.fps_spinbox.setValue(self.current_fps)
        elif not self.got_debug_fps:
            self.update_fps_counter()

        self.main.switch_frame(curr_frame, render_frame=(x[1] for x in frames_futures))

    def _play_next_audio_frame(self) -> None:
        if not self.main.current_output.prepared or not self.current_audio_output:
            return

        next_buffered_frame = self.current_audio_frame + self.settings.playback_buffer_size

        try:
            frame_future = self.play_buffer_audio.pop()
        except BaseException:
            self.play_pause_button.click()
            return

        if next_buffered_frame < self.current_audio_output.total_frames:
            assert self.current_audio_output.vs_output

            self.play_buffer_audio.appendleft(
                self.current_audio_output.vs_output.get_frame_async(int(next_buffered_frame))  # type: ignore
            )

        self.current_audio_output.render_raw_audio_frame(frame_future.result())
        self.current_audio_frame += 1

    def stop(self) -> None:
        import gc
        import logging

        if not self.play_timer.isActive():
            return

        self.play_timer.stop()

        if self.got_debug_fps and self.play_start_time is not None:
            self.play_end_time = perf_counter_ns()
            self.play_end_frame = Frame(self.main.current_output.last_showed_frame)
        if self.main.statusbar.label.text() == 'Playing':
            self.main.statusbar.label.setText('Ready')

        for future in self.play_buffer:
            future[1].add_done_callback(_del_future)

        self.play_buffer.clear()
        del self.play_buffer

        gc.collect(generation=2)

        self.current_audio_output = self.audio_outputs_combobox.currentValue()

        if not self.audio_muted and self.current_audio_output is not None:
            self.stop_audio()

        self.fps_history.clear()
        self.fps_timer.stop()

        if self.play_start_time is not None and self.got_debug_fps:
            time_interval = (self.play_end_time - self.play_start_time) / 1_000_000_000
            frame_interval = self.play_end_frame - self.play_start_frame
            logging.debug(
                f'{time_interval:.3f} s, {frame_interval} frames, {int(frame_interval) / time_interval:.3f} fps'
            )
            self.play_start_time = None

    def stop_audio(self) -> None:
        if self.current_audio_output is None:
            return

        self.current_audio_output.iodevice.reset()

        self.play_timer_audio.stop()

        for future in self.play_buffer_audio:
            future.add_done_callback(_del_future)  # type: ignore

        self.play_buffer_audio.clear()

        self.current_audio_frame = Frame(0)
        self.audio_outputs_combobox.setEnabled(True)

    def seek_to_start(self, checked: bool | None = None) -> None:
        self.stop()
        self.main.current_output.last_showed_frame = Frame(0)

    def seek_to_end(self, checked: bool | None = None) -> None:
        self.stop()
        self.main.current_output.last_showed_frame = self.main.current_output.total_frames - 1

    def seek_offset(self, offset: int) -> None:
        new_pos = self.main.current_output.last_showed_frame + offset

        if not 0 <= new_pos < self.main.current_output.total_frames:
            return

        if self.play_timer.isActive():
            self.stop()

        self.main.switch_frame(new_pos)

    def on_seek_frame_changed(self, frame: Frame | None) -> None:
        if frame is None:
            return
        qt_silent_call(self.seek_time_control.setValue, Time(frame))

    def on_seek_time_changed(self, time: Time | None) -> None:
        if time is None:
            return
        qt_silent_call(self.seek_frame_control.setValue, Frame(time))

    def on_play_pause_clicked(self, checked: bool) -> None:
        if checked:
            self.play()
        else:
            self.stop()

    def on_timeline_clicked(self, frame: Frame, time: Time) -> None:
        if (
            not self.play_timer.isActive() or not self.play_timer_audio.isActive() or self.current_audio_output is None
        ) or self.current_audio_output.vs_output is None:
            return

        self.current_audio_output.iodevice.reset()
        self.current_audio_frame = self.current_audio_output.to_frame(time)

        for future in self.play_buffer_audio:
            future.add_done_callback(lambda future: future.result())

        self.play_buffer_audio.clear()

        for i in range(0, cast(int, self.play_buffer_audio.maxlen)):
            future = self.current_audio_output.vs_output.get_frame_async(  # type: ignore
                int(self.current_audio_frame + Frame(i + 1))
            )
            self.play_buffer_audio.appendleft(future)

    def on_play_n_frames_clicked(self, checked: bool) -> None:
        if checked:
            self.play(Frame(self.main.current_output.last_showed_frame) + Frame(self.seek_frame_control.value()))
        else:
            self.stop()

    def on_fps_changed(self, new_fps: Fraction) -> None:
        if not self.fps_spinbox.isEnabled() or not self.main.current_output:
            return

        self.main.current_output.play_fps = new_fps

        if self.play_timer.isActive():
            self.stop()
            self.play()

    def reset_fps(self, checked: bool | None = None) -> None:
        self.fps_spinbox.setValue(self.main.current_output.fps_num / self.main.current_output.fps_den)

    def on_fps_unlimited_changed(self, state: Qt.CheckState) -> None:
        if state == Qt.CheckState.Checked:
            self.fps_spinbox.setEnabled(False)
            self.fps_reset_button.setEnabled(False)
            self.fps_variable_checkbox.setChecked(False)
            self.fps_variable_checkbox.setEnabled(False)
        elif state == Qt.CheckState.Unchecked:
            self.fps_spinbox.setEnabled(True)
            self.fps_reset_button.setEnabled(True)
            self.fps_spinbox.setValue(float(self.main.current_output.play_fps))
            self.fps_variable_checkbox.setEnabled(True)

        if self.play_timer.isActive():
            self.stop()
            self.play()

    def on_fps_variable_changed(self, state: Qt.CheckState) -> None:
        if state == Qt.CheckState.Checked:
            self.fps_spinbox.setEnabled(False)
            self.fps_reset_button.setEnabled(False)
            self.fps_unlimited_checkbox.setEnabled(False)
            self.fps_unlimited_checkbox.setChecked(False)
        elif state == Qt.CheckState.Unchecked:
            self.fps_spinbox.setEnabled(True)
            self.fps_reset_button.setEnabled(True)
            self.fps_unlimited_checkbox.setEnabled(True)
            self.fps_spinbox.setValue(float(self.main.current_output.play_fps))

        if self.play_timer.isActive():
            self.stop()
            self.play()

    def update_fps_counter(self) -> None:
        if self.fps_spinbox.isEnabled():
            return

        self.fps_history.append(perf_counter_ns())
        if len(self.fps_history) == 1:
            return

        elapsed_total = 0
        for i in range(len(self.fps_history) - 1):
            elapsed_total += self.fps_history[i + 1] - self.fps_history[i]

        self.current_fps = 1_000_000_000 / (elapsed_total / len(self.fps_history))

    def updateMuteGui(self) -> None:
        if self.volume == 0 or self.audio_muted:
            self.mute_button.setText('🔇')
        elif self.volume <= 33:
            self.mute_button.setText('🔈')
        elif self.volume <= 66:
            self.mute_button.setText('🔉')
        else:
            self.mute_button.setText('🔊')

    def on_mute_clicked(self) -> None:
        if not hasattr(self, 'audio_outputs_combobox'):
            return

        if self.volume == 0 and self.audio_muted:
            self.setVolume(50, True)
        else:
            self.setMute(not self.audio_muted)

    def setMute(self, isMuted: bool) -> None:
        self.audio_muted = isMuted

        if not isMuted:
            if self.play_timer.isActive() and not self.play_timer_audio.isActive():
                self.play_audio()
        elif self.play_timer_audio.isActive():
            self.stop_audio()

        self.updateMuteGui()

    def setVolume(self, newVolume: float, updateGui: bool = False) -> None:
        self.volume = newVolume

        self.setMute(self.volume == 0)

        if newVolume:
            for output in self.audio_outputs:
                output.volume = newVolume / 100.0

        if updateGui:
            qt_silent_call(self.audio_volume_slider.setValue, self.volume)

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'seek_interval_frame': self.seek_frame_control.value(),
            'current_audio_output_index': self.audio_outputs_combobox.currentIndex()
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'seek_interval_frame', Frame, self.seek_frame_control.setValue)
        try_load(state, 'audio_outputs', AudioOutputs, self.rescan_outputs)
        try_load(state, 'current_audio_output_index', int, self.audio_outputs_combobox.setCurrentIndex)
        try_load(state, 'audio_muted', bool, self.setMute)
        try_load(state, 'volume', int, partial(self.setVolume, updateGui=True))
        super().__setstate__(state)



from __future__ import annotations

from typing import Any, Sequence, cast

from PyQt6.QtCore import QAbstractTableModel, QModelIndex, QRect, Qt
from PyQt6.QtGui import QBrush, QColor, QPainter, QPaintEvent, QPen
from PyQt6.QtWidgets import QLabel, QStatusBar, QWidget

from ..abstracts import PushButton

__all__ = [
    'StatusBar',
    'Switch',
    'TableModel'
]


class StatusBar(QStatusBar):
    label_names = (
        'total_frames_label', 'duration_label', 'resolution_label',
        'pixel_format_label', 'fps_label', 'frame_props_label', 'label'
    )

    total_frames_label: QLabel
    duration_label: QLabel
    resolution_label: QLabel
    pixel_format_label: QLabel
    fps_label: QLabel
    frame_props_label: QLabel
    label: QLabel

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.permament_start_index = 0

    def addPermanentWidget(self, widget: QWidget, stretch: int = 0) -> None:
        self.insertPermanentWidget(self.permament_start_index, widget, stretch)

    def addWidget(self, widget: QWidget, stretch: int = 0) -> None:
        self.permament_start_index += 1
        super().addWidget(widget, stretch)

    def addWidgets(self, widgets: Sequence[QWidget], stretch: int = 0) -> None:
        for widget in widgets:
            self.addWidget(widget, stretch)

    def insertWidget(self, index: int, widget: QWidget, stretch: int = 0) -> int:
        self.permament_start_index += 1
        return super().insertWidget(index, widget, stretch)


class Switch(PushButton):
    def __init__(
        self, radius: int = 10, width: int = 32, border: int = 1,
        state_texts: tuple[str, str, int] = ('OFF', 'ON', 1),
        state_colors: tuple[QColor, QColor] = (QColor(69, 83, 100), QColor(96, 121, 139)),
        border_color: QColor = QColor(69, 83, 100), text_color: QColor = QColor(224, 225, 226),
        background_color: QColor = QColor(25, 35, 45), **kwargs: Any
    ):
        super().__init__(**kwargs)

        self.radius = radius
        self.sw_width = width
        self.border = border
        self.state_texts = state_texts
        self.state_colors = state_colors
        self.border_color = border_color
        self.text_color = text_color
        self.background_color = background_color

        self.setCheckable(True)
        self.setMinimumWidth((width + border) * 2)
        self.setMinimumHeight((radius + border) * 2)

    def paintEvent(self, event: QPaintEvent) -> None:
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.translate(self.rect().center())
        painter.setBrush(self.background_color)

        pen = QPen(self.border_color)
        pen.setWidth(self.border)
        painter.setPen(pen)

        painter.drawRoundedRect(
            QRect(-self.sw_width, -self.radius, self.sw_width * 2, self.radius * 2), self.radius, self.radius
        )
        painter.setBrush(QBrush(self.state_colors[int(self.isChecked())]))

        switch_rect = QRect(
            -self.radius, -self.radius + self.border, self.sw_width + self.radius, self.radius * 2 - self.border * 2
        )

        if not self.isChecked():
            switch_rect.moveLeft(-self.sw_width)

        painter.setPen(pen)
        painter.drawRoundedRect(switch_rect, self.radius, self.radius)

        textpen = QPen(self.text_color)
        textpen.setWidth(self.state_texts[2])

        painter.setPen(textpen)
        painter.drawText(switch_rect, Qt.AlignmentFlag.AlignCenter, cast(str, self.state_texts[int(self.isChecked())]))


class TableModel(QAbstractTableModel):
    def __init__(self, data: list[list[Any]] = [], columns: list[Any] | bool = True, rows: bool = True) -> None:
        super().__init__()

        self._data = data
        self._columns = columns
        self._rows = rows

    def data(self, index: QModelIndex, role: int) -> Any:
        if role == Qt.ItemDataRole.DisplayRole:
            return self._data[index.row()][index.column()]

    def rowCount(self, index: QModelIndex) -> int:
        return len(self._data)

    def columnCount(self, index: QModelIndex) -> int:
        return len(self._data) and len(self._data[0])

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if isinstance(self._columns, bool):
                    if self._columns:
                        return super().headerData(section, orientation, role)
                else:
                    return str(self._columns[section])

            if orientation == Qt.Vertical:
                if self._rows:
                    return super().headerData(section, orientation, role)



from __future__ import annotations

from typing import Any, Generic, Iterator, Sequence, TypeVar

from PyQt6.QtCore import QAbstractListModel, QModelIndex, Qt

T = TypeVar('T')


__all__ = [
    'GeneralModel'
]


class GeneralModel(QAbstractListModel, Generic[T]):
    __slots__ = ('items',)

    def __class_getitem__(cls, _content_type: type[T]) -> type[GeneralModel[T]]:

        if _content_type is float:
            return _GeneralModel_float
        elif _content_type is int:
            return _GeneralModel_int
        elif _content_type is str:
            return _GeneralModel_string

        class _GenericModel_inner(GeneralModel):  # type: ignore
            content_type = _content_type

        return _GenericModel_inner

    content_type: type[T]
    to_title: bool

    def __init__(self, init_seq: Sequence[T], to_title: bool = True) -> None:
        super().__init__()
        self.items = list(init_seq)
        self.to_title = to_title

    def __getitem__(self, i: int) -> T:
        return self.items[i]

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[T]:
        return iter(self.items)

    def index_of(self, item: T) -> int:
        return self.items.index(item)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.UserRole) -> Any:
        if (not index.isValid() or index.row() >= len(self.items)):
            return None

        if role == Qt.ItemDataRole.DisplayRole:
            value = self.items[index.row()]

            str_value = self._displayValue(value)

            if self.content_type is str and self.to_title:
                return str_value.title()

            return str_value

        if role == Qt.ItemDataRole.UserRole:
            return self.items[index.row()]

        return None

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.items)

    def _displayValue(self, value: T) -> str:
        raise NotImplementedError


class _GeneralModel_float(GeneralModel):  # type: ignore
    content_type = float

    def _displayValue(self, value: float) -> str:
        return str(round(value * 100)) + '%'


class _GeneralModel_int(GeneralModel):  # type: ignore
    content_type = int

    def _displayValue(self, value: int) -> str:
        return str(value)


class _GeneralModel_string(GeneralModel):  # type: ignore
    content_type = str

    def _displayValue(self, value: str) -> str:
        return value



from __future__ import annotations
from typing import TYPE_CHECKING

from ...core import AbstractToolbar, LineEdit, PushButton
from .settings import DebugSettings

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'DebugToolbar'
]


class DebugToolbar(AbstractToolbar):
    __slots__ = ('exec_lineedit', )

    _no_visibility_choice = True

    settings: DebugSettings

    def __init__(self, main: MainWindow) -> None:
        from ...utils import debug

        super().__init__(main, DebugSettings(self))

        self.setup_ui()

        if self.settings.DEBUG_TOOLBAR_BUTTONS_PRINT_STATE:
            self.filter = debug.EventFilter(main)

            self.main.toolbars.main.widget.installEventFilter(self.filter)  # type: ignore

            for toolbar in self.main.toolbars:
                toolbar.widget.installEventFilter(self.filter)  # type: ignore

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.exec_lineedit = LineEdit(
            'Python statement in context of DebugToolbar.exec_button_clicked()',
            self, editingFinished=self.exec_button_clicked
        )

        self.hlayout.addWidgets([
            PushButton('Test', self, clicked=self.test_button_clicked),
            PushButton('Break', self),
            self.exec_lineedit,
            PushButton('Exec', self, clicked=self.exec_button_clicked)
        ])

        self.hlayout.addStretch()

    def test_button_clicked(self, checked: bool | None = None) -> None:
        from vstools.utils.vs_proxy import clear_cache
        clear_cache()

    def exec_button_clicked(self, checked: bool | None = None) -> None:
        try:
            exec(self.exec_lineedit.text())
        except BaseException as e:
            import logging

            logging.error(e)

    def break_button_clicked(self, checked: bool | None = None) -> None:
        breakpoint()



from __future__ import annotations

from typing import Any, Mapping

from .units import Frame
from .yaml import YAMLObjectWrapper

__all__ = [
    'Scene'
]


class Scene(YAMLObjectWrapper):
    __slots__ = ('start', 'end', 'label')

    def __init__(self, start: Frame | None = None, end: Frame | None = None, label: str = '') -> None:
        self.setValue(start, end, label)

    def setValue(self, start: Frame | None, end: Frame | None, label: str) -> None:
        if start is not None and end is not None:
            self.start = start
            self.end = end
        elif start is not None:
            self.start = start
            self.end = start
        elif end is not None:
            self.start = end
            self.end = end
        else:
            raise ValueError

        if self.start > self.end:
            self.start, self.end = self.end, self.start

        self.label = label

    def __str__(self) -> str:
        result = ''

        if self.start == self.end:
            result = f'{self.start}'
        else:
            result = f'{self.start} - {self.end}'

        if self.label != '':
            result += f': {self.label}'

        return result

    def __repr__(self) -> str:
        return f"Scene({self.start}, {self.end}, '{self.label}')"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Scene):
            raise NotImplementedError
        return self.start == other.start and self.end == other.end

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, Scene):
            raise NotImplementedError
        if self.start != other.start:
            return self.start > other.start
        else:
            return self.end > other.end

    def duration(self) -> Frame:
        return self.end - self.start

    def __contains__(self, frame: Frame) -> bool:
        return self.start <= frame <= self.end

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try:
            if not isinstance(state['start'], Frame):
                raise TypeError("Start frame of Scene is not a Frame. It's most probably corrupted.")

            if not isinstance(state['end'], Frame):
                raise TypeError("End frame of Scene is not a Frame. It's most probably corrupted.")

            if not isinstance(state['label'], str):
                raise TypeError("Label of Scene is not a string. It's most probably corrupted.")
        except KeyError:
            raise KeyError(
                "Scene lacks one or more of its fields."
                f"It's most probably corrupted. Check those: {', '.join(self.__slots__)}."
            )

        self.setValue(state['start'], state['end'], state['label'])



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Mapping

from PyQt6.QtCore import QKeyCombination, Qt
from PyQt6.QtWidgets import QComboBox

from ...core import (
    AbstractToolbar, CheckBox, ComboBox, Frame, FrameEdit, PushButton, Time, TimeEdit, VideoOutput, try_load
)
from ...models import VideoOutputs
from ...utils import qt_silent_call

if TYPE_CHECKING:
    from ...main import MainSettings, MainWindow


__all__ = [
    'MainToolbar'
]


class MainToolbar(AbstractToolbar):
    _no_visibility_choice = True
    storable_attrs = ('outputs', )

    __slots__ = (
        *storable_attrs,
        'outputs_combobox', 'frame_control', 'copy_frame_button',
        'time_control', 'copy_timestamp_button',
        'switch_timeline_mode_button', 'settings_button'
    )

    outputs: VideoOutputs | None

    settings: MainSettings

    def __init__(self, main_window: MainWindow) -> None:
        super().__init__(main_window, main_window.settings)
        self.setup_ui()

        self.outputs = None

        self.add_shortcuts()

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.setVisible(True)

        self.outputs_combobox = ComboBox[VideoOutput](
            self, editable=True, insertPolicy=QComboBox.InsertPolicy.InsertAtCurrent,
            duplicatesEnabled=True, sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents
        )
        self.outputs_combobox.currentIndexChanged.connect(self.main.switch_output)
        self.outputs_combobox.view().setMinimumWidth(
            self.outputs_combobox.minimumSizeHint().width()
        )

        self.frame_control = FrameEdit(self, valueChanged=self.main.switch_frame)
        if not self.settings.INSTANT_FRAME_UPDATE:
            self.frame_control.setKeyboardTracking(False)

        self.copy_frame_button = PushButton('⎘', self, clicked=self.on_copy_frame_button_clicked)

        self.time_control = TimeEdit(self, valueChanged=self.main.switch_frame)

        self.copy_timestamp_button = PushButton('⎘', self, clicked=self.on_copy_timestamp_button_clicked)

        self.sync_outputs_checkbox = CheckBox(
            'Sync Outputs', self, checked=self.settings.SYNC_OUTPUTS, clicked=self.on_sync_outputs_clicked
        )

        self.switch_timeline_mode_button = PushButton(
            'Switch Timeline Mode', self, clicked=self.on_switch_timeline_mode_clicked
        )

        self.settings_button = PushButton('Settings', self, clicked=self.main.app_settings.show)

        self.hlayout.addWidgets([
            self.outputs_combobox,
            self.frame_control, self.copy_frame_button,
            self.time_control, self.copy_timestamp_button,
            self.sync_outputs_checkbox,
            self.get_separator(),
            self.main.graphics_view.controls,
            self.switch_timeline_mode_button,
            self.settings_button
        ])

        self.hlayout.addStretch()

    def add_shortcuts(self) -> None:
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_R).toCombined(), self.main.reload_script
        )

        for i, key in enumerate(self.num_keys):
            self.add_shortcut(key, partial(self.main.switch_output, i))
            self.add_shortcut(
                QKeyCombination(Qt.Modifier.CTRL, key).toCombined(), partial(self.main.switch_output, -(i + 1))
            )

        self.add_shortcut(Qt.Key.Key_S, self.sync_outputs_checkbox.click)
        self.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_Tab).toCombined(),
            lambda: self.main.switch_output(self.outputs_combobox.currentIndex() + 1)
        )
        self.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL | Qt.Modifier.SHIFT, Qt.Key.Key_Tab).toCombined(),
            lambda: self.main.switch_output(self.outputs_combobox.currentIndex() - 1)
        )
        self.add_shortcut(Qt.Key.Key_V, self.on_copy_frame_button_clicked)

    def on_sync_outputs_clicked(self, checked: bool | None = None, force_frame: Frame | None = None) -> None:
        if not self.outputs:
            return

        if checked:
            if not force_frame:
                force_frame = self.main.current_output.last_showed_frame

            for output in self.outputs:
                output.last_showed_frame = output.to_frame(
                    self.main.current_output.to_time(force_frame)
                )

    def on_current_frame_changed(self, frame: Frame) -> None:
        qt_silent_call(self.frame_control.setValue, frame)
        qt_silent_call(self.time_control.setValue, Time(frame))

        if self.outputs and len(self.outputs) > 1 and self.sync_outputs_checkbox.isChecked():
            self.on_sync_outputs_clicked(True, force_frame=frame)

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        qt_silent_call(self.outputs_combobox.setCurrentIndex, index)
        qt_silent_call(self.frame_control.setMaximum, self.main.current_output.total_frames - 1)
        qt_silent_call(self.time_control.setMaximum, self.main.current_output.total_time - Frame(1))

        for graphics_view in self.main.graphics_views:
            if graphics_view.autofit:
                graphics_view.setZoom(None)

    def rescan_outputs(self, outputs: VideoOutputs | None = None) -> None:
        self.outputs = outputs if isinstance(outputs, VideoOutputs) else VideoOutputs(self.main)
        self.main.init_outputs()
        self.outputs_combobox.setModel(self.outputs)

    def on_copy_frame_button_clicked(self, checked: bool | None = None) -> None:
        self.main.clipboard.setText(str(self.main.current_output.last_showed_frame))
        self.main.show_message('Current frame number copied to clipboard')

    def on_copy_timestamp_button_clicked(self, checked: bool | None = None) -> None:
        self.main.clipboard.setText(self.time_control.text())
        self.main.show_message('Current timestamp copied to clipboard')

    def on_switch_timeline_mode_clicked(self, checked: bool | None = None) -> None:
        if self.main.timeline.mode == self.main.timeline.Mode.TIME:
            self.main.timeline.mode = self.main.timeline.Mode.FRAME
        elif self.main.timeline.mode == self.main.timeline.Mode.FRAME:
            self.main.timeline.mode = self.main.timeline.Mode.TIME

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'current_output_index': self.outputs_combobox.currentIndex(),
            'sync_outputs': self.sync_outputs_checkbox.isChecked()
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'outputs', VideoOutputs, self.rescan_outputs)
        try_load(state, 'current_output_index', int, self.main.switch_output)
        try_load(state, 'sync_outputs', bool, self.sync_outputs_checkbox.setChecked)



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QColor, QPainter, QPaintEvent
from PyQt6.QtWidgets import QWidget

from .graphicsview import DragEventType, GraphicsView

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'DragNavigator'
]


class DragNavigator(QWidget):
    __slots__ = ()

    is_tracking = False

    contentsH = contentsW = viewportX = viewportY = viewportH = viewportW = 0

    def __init__(self, main: MainWindow, graphics_view: GraphicsView) -> None:
        from ..abstracts import Timer

        super().__init__(graphics_view)

        self.main = main
        self.graphics_view = graphics_view
        self.graphics_view.dragEvent.connect(self.on_drag)
        self.setAttribute(Qt.WidgetAttribute.WA_NoSystemBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)
        rate = self.main.settings.base_ppi / 96
        self.setGeometry(round(10 * rate), round(10 * rate), round(120 * rate), round(120 * rate))
        self.repaint_timer = Timer(
            timeout=self.repaint_timeout, timerType=Qt.TimerType.PreciseTimer,
            interval=self.main.settings.dragnavigator_timeout
        )
        self.graphics_view.verticalScrollBar().valueChanged.connect(partial(self.on_drag, DragEventType.repaint))
        self.graphics_view.horizontalScrollBar().valueChanged.connect(partial(self.on_drag, DragEventType.repaint))

    def on_drag(self, event_type: DragEventType) -> None:
        # while reloading and moving mouse
        if not hasattr(self.main, 'current_output') or not self.main.current_output:
            return

        if event_type == DragEventType.repaint:
            self.repaint_timer.stop()
            self.repaint_timer.start()
        elif event_type == DragEventType.move and not self.is_tracking:
            return
        elif event_type == DragEventType.start:
            self.is_tracking = True
        elif event_type == DragEventType.stop:
            self.is_tracking = False
            self.setVisible(False)
            return

        scrollbarW = self.graphics_view.horizontalScrollBar()
        scrollbarH = self.graphics_view.verticalScrollBar()

        if not any({scrollbarW.isVisible(), scrollbarH.isVisible()}):
            return self.hide()

        self.setVisible(True)
        self.draw(
            self.main.current_output.width,
            self.main.current_output.height,
            int(scrollbarW.value() / self.graphics_view.currentZoom),
            int(scrollbarH.value() / self.graphics_view.currentZoom),
            int(self.graphics_view.width() / self.graphics_view.currentZoom),
            int(self.graphics_view.height() / self.graphics_view.currentZoom)
        )

    def draw(
        self, contentsW: int, contentsH: int, viewportX: int, viewportY: int, viewportW: int, viewportH: int
    ) -> None:
        self.contentsW = contentsW
        self.contentsH = contentsH
        self.viewportW = min(viewportW, contentsW)
        self.viewportH = min(viewportH, contentsH)
        self.viewportX = min(viewportX, contentsW)
        self.viewportY = min(viewportY, contentsH)

        self.repaint()

    def repaint_timeout(self) -> None:
        self.hide()
        self.repaint_timer.stop()

    def paintEvent(self, event: QPaintEvent) -> None:
        if (self.contentsW == 0) or (self.contentsH == 0) or (self.viewportW == 0) or \
                (self.viewportH == 0) or (self.viewportX >= self.contentsW) or (self.viewportY >= self.contentsH):
            event.ignore()
            return

        norm = 100.0 / max(self.contentsW, self.contentsH, self.viewportW, self.viewportH)

        normContentsWidth, normContentsHeight = round(self.contentsW * norm), round(self.contentsH * norm)
        normViewportWidth, normViewportHeight = round(self.viewportW * norm), round(self.viewportH * norm)
        normVieportX = min(round(self.viewportX * norm), normContentsWidth - normViewportWidth)
        normViwportY = min(round(self.viewportY * norm), normContentsHeight - normViewportHeight)

        cX1 = cY1 = 0
        cX2, cY2 = normContentsWidth - 1, normContentsHeight - 1

        vX1, vY1 = normVieportX, normViwportY
        vX2, vY2 = normVieportX + normViewportWidth - 1, normViwportY + normViewportHeight - 1

        painter = QPainter(self)

        painter.setPen(QColor(255, 0, 255))
        painter.drawLine(cX1, cY1, cX2, cY1)
        painter.drawLine(cX2, cY1, cX2, cY2)
        painter.drawLine(cX2, cY2, cX1, cY2)
        painter.drawLine(cX1, cY2, cX1, cY1)

        painter.setPen(QColor(0, 255, 0))
        painter.drawLine(vX1, vY1, vX2, vY1)
        painter.drawLine(vX2, vY1, vX2, vY2)
        painter.drawLine(vX2, vY2, vX1, vY2)
        painter.drawLine(vX1, vY2, vX1, vY1)

        event.accept()



from .combobox import *  # noqa: F401, F403
from .dragnavigator import *  # noqa: F401, F403
from .edits import *  # noqa: F401, F403
from .graphicsview import *  # noqa: F401, F403
from .misc import *  # noqa: F401, F403
from .notch import *  # noqa: F401, F403
from .plotting import *  # noqa: F401, F403



from __future__ import annotations

from PyQt6.QtWidgets import QLabel

from ...core import AbstractToolbarSettings, CheckBox, HBoxLayout, LineEdit

__all__ = [
    'SceningSettings'
]


class SceningSettings(AbstractToolbarSettings):
    __slots__ = ('export_template_lineedit', 'always_show_scene_marks_checkbox')

    def setup_ui(self) -> None:
        super().setup_ui()

        self.export_template_lineedit = LineEdit('Export Template')

        self.always_show_scene_marks_checkbox = CheckBox('Always show scene marks')

        HBoxLayout(self.vlayout, [QLabel('Default Export Template'), self.export_template_lineedit])

        HBoxLayout(self.vlayout, self.always_show_scene_marks_checkbox)

    def set_defaults(self) -> None:
        self.export_template_lineedit.setText(r'({start},{end}),')
        self.always_show_scene_marks_checkbox.setChecked(False)

    @property
    def default_export_template(self) -> str:
        return self.export_template_lineedit.text()

    @property
    def always_show_scene_marks(self) -> bool:
        return self.always_show_scene_marks_checkbox.isChecked()




from __future__ import annotations

from fractions import Fraction
from pathlib import Path
from typing import Any

from vapoursynth import AudioNode, RawNode, VideoNode

from .info import is_preview

__all__ = [
    'set_timecodes',

    'update_node_info'
]


def set_timecodes(
    index: int, timecodes: str | Path | dict[
        tuple[int | None, int | None], float | tuple[int, int] | Fraction
    ] | list[Fraction], node: VideoNode | None = None, den: int = 1001
) -> None:
    if is_preview():
        from ..core import main_window

        if isinstance(node, VideoNode) and node.fps_den and node.fps_num:
            den = node.fps_den

        main_window().update_timecodes_info(index, timecodes, den)


def update_node_info(node_type: type[RawNode] | type[VideoNode] | type[AudioNode], index: int, **kwargs: Any) -> None:
    if is_preview():
        from ..core import main_window
        main_window().set_node_info(node_type, index, **kwargs)



from __future__ import annotations

from ...core import AbstractToolbarSettings

__all__ = [
    'PipetteSettings'
]


class PipetteSettings(AbstractToolbarSettings):
    __slots__ = ()

    _add_to_tab = False

    def setup_ui(self) -> None:
        super().setup_ui()

    def set_defaults(self) -> None:
        pass



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, NamedTuple

from PyQt6.QtWidgets import QSizePolicy

from ..core import ExtendedWidgetBase, Frame, NotchProvider

if TYPE_CHECKING:
    from ..main import MainWindow


__all__ = [
    'AbstractPlugin',
    'PluginConfig'
]


class PluginConfig(NamedTuple):
    namespace: str
    display_name: str
    visible_in_tab: bool = True


class AbstractPlugin(ExtendedWidgetBase, NotchProvider):
    _config: ClassVar[PluginConfig]

    index: int = -1

    def __init__(self, main: MainWindow) -> None:
        try:
            super().__init__(main)
            self.init_notches(main)
        except TypeError as e:
            print('\tMake sure you\'re inheriting a QWidget!\n')

            raise e

        self.main = main

        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)

        self.setup_ui()

        self.add_shortcuts()

        self.set_qobject_names()

    def init_outputs(self) -> None:
        ...

    def add_shortcuts(self) -> None:
        ...

    def on_current_frame_changed(self, frame: Frame) -> None:
        ...

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        ...

    @property
    def is_notches_visible(self) -> bool:
        return (not self._config.visible_in_tab) or self.index == self.main.plugins.plugins_tab.currentIndex()



from __future__ import annotations

from dataclasses import dataclass

import vapoursynth as vs

__all__ = [
    'CroppingInfo',
    'VideoOutputNode',
    'Stretch'
]


@dataclass
class CroppingInfo:
    top: int
    left: int
    width: int
    height: int
    active: bool = True
    is_absolute: bool = False


@dataclass
class VideoOutputNode:
    clip: vs.VideoNode
    alpha: vs.VideoNode | None


@dataclass
class Stretch:
    amount: int = 0



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtCore import QItemSelection, QItemSelectionModel, QModelIndex, Qt, QTimer
from PyQt6.QtWidgets import QTableView

from ...core import (
    ExtendedDialog, ExtendedTableView, Frame, FrameEdit, HBoxLayout, LineEdit, PushButton, Time, TimeEdit, VBoxLayout
)
from ...models import SceningList
from ...utils import qt_silent_call

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'SceningListDialog'
]


class SceningListDialog(ExtendedDialog):
    __slots__ = (
        'main', 'scening_list',
        'name_lineedit', 'tableview',
        'start_frame_control', 'end_frame_control',
        'start_time_control', 'end_time_control',
        'label_lineedit',
    )

    def __init__(self, main: MainWindow) -> None:
        super().__init__(main)

        self.main = main
        self.scening_list = SceningList()

        self.setWindowTitle('Scening List View')
        self.setup_ui()

        self.end_frame_control.valueChanged.connect(self.on_end_frame_changed)
        self.end_time_control.valueChanged.connect(self.on_end_time_changed)
        self.label_lineedit.textChanged.connect(self.on_label_changed)
        self.name_lineedit.textChanged.connect(self.on_name_changed)
        self.start_frame_control.valueChanged.connect(self.on_start_frame_changed)
        self.start_time_control.valueChanged.connect(self.on_start_time_changed)
        self.tableview.doubleClicked.connect(self.on_tableview_clicked)
        self.delete_button.clicked.connect(self.on_delete_clicked)

        self.tableview.add_shortcut(Qt.Key.Key_Delete, self.delete_button.click)
        self.set_qobject_names()

    def setup_ui(self) -> None:
        self.name_lineedit = LineEdit('Scening list name')

        self.tableview = ExtendedTableView()
        self.tableview.setSelectionMode(QTableView.SelectionMode.SingleSelection)
        self.tableview.setSelectionBehavior(QTableView.SelectionBehavior.SelectRows)
        self.tableview.setSizeAdjustPolicy(QTableView.SizeAdjustPolicy.AdjustToContents)

        self.start_frame_control = FrameEdit()
        self.end_frame_control = FrameEdit()

        self.start_time_control = TimeEdit()
        self.end_time_control = TimeEdit()

        self.label_lineedit = LineEdit(placeholder='Label')

        self.delete_button = PushButton('Delete Selected Scene', enabled=False)
        self.delete_button.setAutoDefault(False)

        VBoxLayout(self, [
            self.name_lineedit, self.tableview
        ]).addLayout(
            HBoxLayout([
                self.start_frame_control,
                self.end_frame_control,
                self.start_time_control,
                self.end_time_control,
                self.label_lineedit,
                self.delete_button
            ])
        )

    def on_current_frame_changed(self, frame: Frame, time: Time) -> None:
        if not self.isVisible():
            return
        if self.tableview.selectionModel() is None:
            return
        selection = QItemSelection()
        for i, scene in enumerate(self.scening_list):
            if frame in scene:
                index = self.scening_list.index(i, 0)
                selection.select(index, index)
        self.tableview.selectionModel().select(
            selection,
            QItemSelectionModel.SelectionFlag.Rows
            | QItemSelectionModel.SelectionFlag.ClearAndSelect
        )

    def on_current_list_changed(self, scening_list: SceningList) -> None:
        self.scening_list = scening_list

        self.scening_list.rowsMoved.connect(self.on_tableview_rows_moved)

        self.name_lineedit.setText(self.scening_list.name)

        self.tableview.setModel(self.scening_list)
        self.tableview.resizeColumnsToContents()
        self.tableview.selectionModel().selectionChanged.connect(self.on_tableview_selection_changed)
        self.delete_button.setEnabled(False)

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        self.start_frame_control.setMaximum(self.main.current_output.total_frames - 1)
        self.end_frame_control.setMaximum(self.main.current_output.total_frames - 1)
        self.start_time_control.setMaximum(self.main.current_output.total_time - Frame(1))
        self.end_time_control.setMaximum(self.main.current_output.total_time - Frame(1))

    def on_delete_clicked(self, checked: bool | None = None) -> None:
        for model_index in self.tableview.selectionModel().selectedRows():
            self.scening_list.remove(model_index.row())

    def on_end_frame_changed(self, value: Frame | int) -> None:
        if self.tableview.selectionModel() is None:
            return

        frame = Frame(value)

        try:
            index = self.tableview.selectionModel().selectedRows()[0]
        except IndexError:
            return
        if not index.isValid():
            return
        index = index.siblingAtColumn(SceningList.END_FRAME_COLUMN)
        if not index.isValid():
            return
        self.scening_list.setData(index, frame, Qt.ItemDataRole.UserRole)

    def on_end_time_changed(self, time: Time) -> None:
        if self.tableview.selectionModel() is None:
            return
        try:
            index = self.tableview.selectionModel().selectedRows()[0]
        except IndexError:
            return
        if not index.isValid():
            return
        index = index.siblingAtColumn(SceningList.END_TIME_COLUMN)
        if not index.isValid():
            return
        self.scening_list.setData(index, time, Qt.ItemDataRole.UserRole)

    def on_label_changed(self, text: str) -> None:
        if self.tableview.selectionModel() is None:
            return
        try:
            index = self.tableview.selectionModel().selectedRows()[0]
        except IndexError:
            return
        if not index.isValid():
            return
        index = self.scening_list.index(index.row(), SceningList.LABEL_COLUMN)
        if not index.isValid():
            return
        self.scening_list.setData(index, text, Qt.ItemDataRole.UserRole)

    def on_name_changed(self, text: str) -> None:
        assert hasattr(self.main.toolbars, 'scening')

        i = self.main.toolbars.scening.lists.index_of(self.scening_list)
        index = self.main.toolbars.scening.lists.index(i)
        self.main.toolbars.scening.lists.setData(index, text, Qt.ItemDataRole.UserRole)

    def on_start_frame_changed(self, value: Frame | int) -> None:
        frame = Frame(value)
        try:
            index = self.tableview.selectionModel().selectedRows()[0]
        except IndexError:
            return
        if not index.isValid():
            return
        index = index.siblingAtColumn(SceningList.START_FRAME_COLUMN)
        if not index.isValid():
            return
        self.scening_list.setData(index, frame, Qt.ItemDataRole.UserRole)

    def on_start_time_changed(self, time: Time) -> None:
        try:
            index = self.tableview.selectionModel().selectedRows()[0]
        except IndexError:
            return
        if not index.isValid():
            return
        index = index.siblingAtColumn(SceningList.START_TIME_COLUMN)
        if not index.isValid():
            return
        self.scening_list.setData(index, time, Qt.ItemDataRole.UserRole)

    def on_tableview_clicked(self, index: QModelIndex) -> None:
        if index.column() in {SceningList.START_FRAME_COLUMN, SceningList.END_FRAME_COLUMN}:
            self.main.switch_frame(Frame(self.scening_list.data(index)))
        if index.column() == SceningList.START_TIME_COLUMN:
            self.main.switch_frame(Frame(self.scening_list.data(index.siblingAtColumn(SceningList.START_FRAME_COLUMN))))
        if index.column() == SceningList.END_TIME_COLUMN:
            self.main.switch_frame(Frame(self.scening_list.data(index.siblingAtColumn(SceningList.END_FRAME_COLUMN))))

    def on_tableview_rows_moved(
        self, parent_index: QModelIndex, start_i: int, end_i: int, dest_index: QModelIndex, dest_i: int
    ) -> None:
        QTimer.singleShot(0, lambda: self.tableview.selectRow(dest_i))

    def on_tableview_selection_changed(self, selected: QItemSelection, deselected: QItemSelection) -> None:
        if len(selected.indexes()) == 0:
            self.delete_button.setEnabled(False)
            self.start_frame_control.setEnabled(False)
            self.end_frame_control.setEnabled(False)
            self.start_time_control.setEnabled(False)
            self.end_time_control.setEnabled(False)
            self.label_lineedit.setEnabled(False)
            return

        index = selected.indexes()[0]
        scene = self.scening_list[index.row()]

        qt_silent_call(self.start_frame_control.setValue, scene.start)
        qt_silent_call(self.end_frame_control.setValue, scene.end)
        qt_silent_call(self.start_time_control.setValue, Time(scene.start))
        qt_silent_call(self.end_time_control.setValue, Time(scene.end))
        qt_silent_call(self.label_lineedit.setText, scene.label)

        self.delete_button.setEnabled(True)
        self.start_frame_control.setEnabled(True)
        self.end_frame_control.setEnabled(True)
        self.start_time_control.setEnabled(True)
        self.end_time_control.setEnabled(True)
        self.label_lineedit.setEnabled(True)



from __future__ import annotations

from collections import deque
from copy import deepcopy
from time import perf_counter
from typing import TYPE_CHECKING

import vapoursynth as vs
from PyQt6.QtCore import QMetaObject, Qt
from PyQt6.QtWidgets import QLabel

from ...core import AbstractToolbar, CheckBox, Frame, PushButton, Time, Timer
from ...core.custom import FrameEdit
from ...utils import qt_silent_call, strfdelta
from .settings import BenchmarkSettings


if TYPE_CHECKING:
    from ...main import MainWindow
    from vapoursynth import _Future as Future
else:
    from concurrent.futures import Future


__all__ = [
    'BenchmarkToolbar'
]


class BenchmarkToolbar(AbstractToolbar):
    __slots__ = (
        'start_frame_control',
        'end_frame_control', 'total_frames_control',
        'prefetch_checkbox', 'unsequenced_checkbox',
        'run_abort_button', 'info_label', 'running',
        'unsequenced', 'run_start_time', 'start_frame',
        'end_frame', 'total_frames', 'frames_left', 'buffer',
        'update_info_timer', 'sequenced_timer'
    )

    settings: BenchmarkSettings

    def __init__(self, main: MainWindow) -> None:
        super().__init__(main, BenchmarkSettings(self))

        self.setup_ui()

        self.running = False
        self.unsequenced = False
        self.buffer = deque[Future[vs.VideoFrame]]()
        self.run_start_time = 0.0
        self.start_frame = Frame(0)
        self.end_frame = Frame(0)
        self.total_frames = Frame(0)
        self.frames_left = Frame(0)

        self.sequenced_timer = Timer(
            timeout=self._request_next_frame_sequenced, timerType=Qt.TimerType.PreciseTimer, interval=0
        )

        self.update_info_timer = Timer(timeout=self.update_info, timerType=Qt.TimerType.PreciseTimer)

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.start_frame_control = FrameEdit(self, valueChanged=lambda value: self.update_controls(start=value))

        self.end_frame_control = FrameEdit(self, valueChanged=lambda value: self.update_controls(end=value))

        self.total_frames_control = FrameEdit(self, 1, valueChanged=lambda value: self.update_controls(total=value))

        self.unsequenced_checkbox = CheckBox(
            'Unsequenced', self, checked=True, tooltip=(
                "If enabled, next frame will be requested each time frameserver returns completed frame.\n"
                "If disabled, first frame that's currently processing will be waited before requesting the next one."
            )
        )

        self.prefetch_checkbox = CheckBox(
            'Prefetch', self, checked=True, tooltip='Request multiple frames in advance.',
            stateChanged=self.on_prefetch_changed
        )

        self.run_abort_button = PushButton('Run', self, checkable=True, clicked=self.on_run_abort_pressed)

        self.info_label = QLabel(self)

        self.hlayout.addWidgets([
            QLabel('Start:'), self.start_frame_control,
            QLabel('End:'), self.end_frame_control,
            QLabel('Total:'), self.total_frames_control,
            self.prefetch_checkbox,
            self.unsequenced_checkbox,
            self.run_abort_button,
            self.info_label,
        ])
        self.hlayout.addStretch()

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        self.start_frame_control.setMaximum(self.main.current_output.total_frames - 1)
        self.end_frame_control.setMaximum(self.main.current_output.total_frames - 1)
        self.total_frames_control.setMaximum(self.main.current_output.total_frames - Frame(1))

    def run(self) -> None:
        if self.settings.clear_cache_enabled:
            from vstools.utils.vs_proxy import clear_cache
            clear_cache()

        if self.settings.frame_data_sharing_fix_enabled:
            self.main.current_output.update_graphic_item(
                self.main.current_scene.pixmap().copy(),
                graphics_scene_item=self.main.current_output.graphics_scene_item
            )

        self.start_frame = self.start_frame_control.value()
        self.end_frame = self.end_frame_control.value()
        self.total_frames = self.total_frames_control.value()
        self.frames_left = deepcopy(self.total_frames)
        if self.prefetch_checkbox.isChecked():
            concurrent_requests_count = self.main.settings.usable_cpus_count
        else:
            concurrent_requests_count = 1

        self.unsequenced = self.unsequenced_checkbox.isChecked()
        if not self.unsequenced:
            self.buffer = deque([], concurrent_requests_count)
            self.sequenced_timer.start()

        self.running = True
        self.run_start_time = perf_counter()

        for offset in range(min(int(self.frames_left), concurrent_requests_count)):
            if self.unsequenced:
                self._request_next_frame_unsequenced()
            else:
                frame = self.start_frame + Frame(offset)
                future = self.main.current_output.prepared.clip.get_frame_async(int(frame))
                self.buffer.appendleft(future)

        self.update_info_timer.setInterval(round(float(self.settings.refresh_interval) * 1000))
        self.update_info_timer.start()

    def abort(self) -> None:
        if self.running:
            self.update_info()

        self.running = False
        QMetaObject.invokeMethod(self.update_info_timer, 'stop', Qt.ConnectionType.QueuedConnection)

        if self.run_abort_button.isChecked():
            self.run_abort_button.click()

    def _request_next_frame_sequenced(self) -> None:
        if self.frames_left <= Frame(0):
            self.abort()
            return

        self.buffer.pop().result()

        next_frame = self.end_frame + Frame(1) - self.frames_left
        if next_frame <= self.end_frame:
            new_future = self.main.current_output.prepared.clip.get_frame_async(int(next_frame))
            self.buffer.appendleft(new_future)

        self.frames_left -= Frame(1)

    def _request_next_frame_unsequenced(self, future: Future[vs.VideoFrame] | None = None) -> None:
        if self.frames_left <= Frame(0):
            self.abort()
            return

        if self.running:
            next_frame = self.end_frame + Frame(1) - self.frames_left
            new_future = self.main.current_output.prepared.clip.get_frame_async(int(next_frame))
            new_future.add_done_callback(self._request_next_frame_unsequenced)  # type: ignore

        if future is not None:
            future.result()
        self.frames_left -= Frame(1)

    def on_run_abort_pressed(self, checked: bool) -> None:
        self.set_ui_editable(not checked)
        if checked:
            self.run()
        else:
            self.abort()

    def on_prefetch_changed(self, new_state: Qt.CheckState) -> None:
        if new_state == Qt.CheckState.Checked:
            self.unsequenced_checkbox.setEnabled(True)
        elif new_state == Qt.CheckState.Unchecked:
            self.unsequenced_checkbox.setChecked(False)
            self.unsequenced_checkbox.setEnabled(False)

    def set_ui_editable(self, new_state: bool) -> None:
        self. start_frame_control.setEnabled(new_state)
        self.end_frame_control.setEnabled(new_state)
        self.total_frames_control.setEnabled(new_state)
        self.prefetch_checkbox.setEnabled(new_state)
        self. unsequenced_checkbox.setEnabled(new_state)

    def update_controls(
        self, start: Frame | None = None, end: Frame | None = None, total: Frame | None = None
    ) -> None:
        if not hasattr(self.main, 'current_output'):
            return

        if start is not None:
            end = self.end_frame_control.value()
            total = self.total_frames_control.value()

            if start > end:
                end = start
            total = end - start + Frame(1)
        elif end is not None:
            start = self.start_frame_control.value()
            total = self.total_frames_control.value()

            if end < start:
                start = end
            total = end - start + Frame(1)
        elif total is not None:
            start = self.start_frame_control.value()
            end = self.end_frame_control.value()
            old_total = end - start + Frame(1)
            delta = total - old_total

            end += delta
            if end > (e := self.main.current_output.total_frames - 1):
                start -= end - e
                end = e
        else:
            return

        qt_silent_call(self.start_frame_control.setValue, start)
        qt_silent_call(self.end_frame_control.setValue, end)
        qt_silent_call(self.total_frames_control.setValue, total)

    def update_info(self) -> None:
        run_time = Time(seconds=(perf_counter() - self.run_start_time))
        frames_done = self.total_frames - self.frames_left
        fps = int(frames_done) / float(run_time)

        info_str = (f"{frames_done}/{self.total_frames} frames in {strfdelta(run_time, '%M:%S.%Z')}, {fps:.4f} fps")
        self.info_label.setText(info_str)



from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import Iterable

from ..main import MainWindow

BASE_URL = 'https://api.github.com/repos'
PLUGINS_PATH = 'Irrational-Encoding-Wizardry/vs-preview-plugins'
BRANCH = 'master'

CONTENT_URL = f'https://raw.githubusercontent.com/{PLUGINS_PATH}/{BRANCH}/{{path}}'

plugins_commands = ('install', 'uninstall', 'update')


def get_plugins() -> dict[str, Iterable[str]]:
    from requests import get

    response = get(f'{BASE_URL}/{PLUGINS_PATH}/git/trees/{BRANCH}?recursive=1')

    if response.status_code != 200:
        logging.error('There was an error fetching plugin data!')
        sys.exit(1)

    data = response.json()

    if data['truncated']:
        logging.error('Bug setsu to fix!')
        sys.exit(1)

    data = data['tree']

    base_paths = list[str](set(d['path'] for d in data if d['type'] == 'tree' and '/' not in d['path']))

    def _get_generator(path: str) -> Iterable[str]:
        return (
            d['path'] for d in data
            if ('type' not in d or d['type'] != 'tree') and d['path'].startswith(f'{path}/')
        )

    return {path: _get_generator(path) for path in base_paths}


def install_plugins(plugins: list[str], force: bool = False, no_deps: bool = False) -> None:
    from shutil import copytree, rmtree
    from subprocess import Popen
    from tempfile import TemporaryDirectory

    from requests import Session

    existing_plugins = get_plugins()

    found_plugins = set(existing_plugins.keys()).intersection(plugins)

    not_found_plugins = set(plugins) - found_plugins
    if not_found_plugins:
        logging.warn(f'Not found the following plugins: "{", ".join(not_found_plugins)}"')

    with Session() as s:
        for plugin in found_plugins:
            if (MainWindow.global_plugins_dir / plugin).exists():
                if not force:
                    logging.info(f'Skipping "{plugin}" as it\'s already installed.')
                    continue
                else:
                    rmtree(MainWindow.global_plugins_dir / plugin)

            logging.info(f'Downloading "{plugin}"...')

            with TemporaryDirectory() as tempdir:
                tempdir = Path(tempdir)

                requirements = list[Path]()

                for file in existing_plugins[plugin]:
                    logging.info(f'Collecting "{file}"...')

                    temp = tempdir / file
                    (temp if temp.is_dir() else temp.parent).mkdir(parents=True, exist_ok=True)

                    with s.get(CONTENT_URL.format(path=file), stream=True) as req:
                        req.raise_for_status()
                        with temp.open('wb') as f:
                            for chunk in req.iter_content(8192):
                                f.write(chunk)

                    if Path(file).name == 'requirements.txt':
                        requirements.append(temp)

                for requirement in requirements:
                    if not no_deps:
                        Popen([sys.executable, '-m', 'pip', 'install', '-r', str(requirement)]).wait()
                    requirement.unlink()

                copytree(tempdir / plugin, MainWindow.global_plugins_dir / plugin)


def uninstall_plugins(plugins: list[str], ignore: bool = False) -> None:
    from shutil import rmtree

    not_found_plugins = set[str]()

    for plugin in plugins:
        if (MainWindow.global_plugins_dir / plugin).exists():
            rmtree(MainWindow.global_plugins_dir / plugin)
        else:
            not_found_plugins.add(plugin)

    if ignore:
        return

    found_plugins = set(plugins) - not_found_plugins

    if not_found_plugins:
        logging.warn(f'Not found the following plugins: "{", ".join(not_found_plugins)}"')

    logging.info(f'Successfully uninstalled "{", ".join(found_plugins)}"')



from __future__ import annotations

import sys
from functools import partial, wraps
from string import Template
from typing import TYPE_CHECKING, Any, Callable

from PyQt6.QtCore import QSignalBlocker

if TYPE_CHECKING:
    from vstools import F, P, R, T

from ..core import Time, main_window

__all__ = [
    'exit_func',
    'qt_silent_call',
    'strfdelta',
    'fire_and_forget',
    'set_status_label'
]


def exit_func(ret_code: int, no_exit: bool) -> int:
    if not no_exit:
        sys.exit(ret_code)

    return ret_code


# it is a BuiltinMethodType at the same time
def qt_silent_call(qt_method: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> R:
    block = QSignalBlocker(qt_method.__self__)  # type: ignore
    ret = qt_method(*args, **kwargs)
    del block
    return ret


class DeltaTemplate(Template):
    delimiter = '%'


def strfdelta(time: Time, output_format: str) -> str:
    seconds = time.value.seconds % (24 * 3600)
    hours = seconds // 3600
    seconds %= 3600
    minutes = seconds // 60
    seconds %= 60
    milliseconds = time.value.microseconds // 1000

    template = DeltaTemplate(output_format)

    return template.substitute(
        D='{:d}'.format(time.value.days),
        H='{:02d}'.format(hours),
        M='{:02d}'.format(minutes),
        S='{:02d}'.format(seconds),
        Z='{:03d}'.format(milliseconds),
        h='{:d}'.format(hours),
        m='{:2d}'.format(minutes),
        s='{:2d}'.format(seconds)
    )


def fire_and_forget(f: F) -> F:
    from asyncio import get_event_loop_policy, get_running_loop

    @wraps(f)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        try:
            loop = get_running_loop()
        except RuntimeError:
            loop = get_event_loop_policy().get_event_loop()
        return loop.run_in_executor(None, partial(f, *args, **kwargs))

    return wrapped  # type: ignore


def set_status_label(label: str) -> Callable[[F], F]:
    def _decorator(func: Callable[..., T]) -> Any:
        @wraps(func)
        def _wrapped(*args: Any, **kwargs: Any) -> T:
            main = main_window()

            main.statusbar.label.setText(label)

            ret = func(*args, **kwargs)

            main.statusbar.label.setText('Ready')

            return ret
        return _wrapped

    return _decorator



from __future__ import annotations

from ...core import AbstractToolbarSettings

__all__ = [
    'DebugSettings'
]


class DebugSettings(AbstractToolbarSettings):
    __slots__ = ()

    _add_to_tab = False

    DEBUG_PLAY_FPS = False
    DEBUG_TOOLBAR = False
    DEBUG_TOOLBAR_BUTTONS_PRINT_STATE = False

    def setup_ui(self) -> None:
        super().setup_ui()

    def set_defaults(self) -> None:
        pass



from __future__ import annotations

from abc import ABCMeta
from typing import TYPE_CHECKING, Any, cast, no_type_check

from PyQt6 import sip
from yaml import YAMLObject, YAMLObjectMetaclass

if TYPE_CHECKING:
    from vstools import T


__all__ = [
    'AbstractYAMLObjectSingleton',
    'QABC',
    'QYAMLObject',
    'QYAMLObjectSingleton',
    'QAbstractYAMLObjectSingleton'
]


class SingletonMeta(type):
    def __init__(cls: type[T], name: str, bases: tuple[type, ...], dct: dict[str, Any]) -> None:
        super().__init__(name, bases, dct)  # type: ignore
        cls.instance: T | None = None  # type: ignore

    def __call__(cls, *args: Any, **kwargs: Any) -> T:  # type: ignore
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance

    def __new__(cls: type[type], name: str, bases: tuple[type, ...], dct: dict[str, Any]) -> type:
        subcls = super(SingletonMeta, cls).__new__(cls, name, bases, dct)  # type: ignore
        singleton_new = None
        for entry in subcls.__mro__:
            if entry.__class__ is SingletonMeta:
                singleton_new = entry.__new__
        if subcls.__new__ is not singleton_new:
            subcls.__default_new__ = subcls.__new__
            subcls.__new__ = singleton_new
        return cast(type, subcls)


class Singleton(metaclass=SingletonMeta):
    @no_type_check
    def __new__(cls: type[T], *args: Any, **kwargs: Any) -> T:
        if cls.instance is None:
            if hasattr(cls, '__default_new__'):
                cls.instance = cls.__default_new__(cls, *args, **kwargs)
            else:
                cls.instance = super(Singleton, cls).__new__(cls)
        return cls.instance


class AbstractYAMLObjectMeta(YAMLObjectMetaclass, ABCMeta):
    pass


class AbstractYAMLObjectMetaClass(YAMLObject, metaclass=AbstractYAMLObjectMeta):
    pass


class AbstractYAMLObjectSingletonMeta(SingletonMeta, AbstractYAMLObjectMeta):
    pass


class AbstractYAMLObjectSingleton(AbstractYAMLObjectMetaClass, Singleton, metaclass=AbstractYAMLObjectSingletonMeta):
    pass


class QABCMeta(sip.wrappertype, ABCMeta):
    pass


class QABC(metaclass=QABCMeta):
    pass


class QSingletonMeta(SingletonMeta, sip.wrappertype):
    pass


class QYAMLObjectMeta(YAMLObjectMetaclass, sip.wrappertype):
    pass


class QYAMLObject(YAMLObject, metaclass=QYAMLObjectMeta):
    pass


class QYAMLObjectSingletonMeta(QSingletonMeta, QYAMLObjectMeta):
    pass


class QYAMLObjectSingleton(QYAMLObject, Singleton, metaclass=QYAMLObjectSingletonMeta):
    pass


class QAbstractYAMLObjectSingletonMeta(QYAMLObjectSingletonMeta):
    pass


class QAbstractYAMLObjectSingleton(QYAMLObjectSingleton, metaclass=QAbstractYAMLObjectSingletonMeta):
    pass



from __future__ import annotations

from colorsys import rgb_to_hls, rgb_to_hsv
from math import ceil, floor, log
from typing import TYPE_CHECKING, Generator, cast
from weakref import WeakKeyDictionary

import vapoursynth as vs
from PyQt6.QtCore import QPoint, Qt
from PyQt6.QtGui import QFont, QMouseEvent
from PyQt6.QtWidgets import QFrame, QGraphicsView, QLabel

from ...core import AbstractToolbar, Frame, PushButton, VideoOutput
from .colorview import ColorView
from .settings import PipetteSettings

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'PipetteToolbar'
]


class PipetteToolbar(AbstractToolbar):
    labels = [
        'position',
        'rgb_label', 'rgb_hex', 'rgb_dec', 'rgb_norm',
        'src_label', 'src_hex', 'src_dec', 'src_norm'
    ]

    __slots__ = (
        'color_view', 'outputs', 'tracking',
        'src_dec_fmt', 'src_hex_fmt', *labels,
        'copy_position_button'
    )

    settings: PipetteSettings

    def __init__(self, main: MainWindow) -> None:
        import ctypes

        super().__init__(main, PipetteSettings(self))

        self.setup_ui()
        self.src_max_val: float = 2**8 - 1
        self.pos_fmt = self.src_hex_fmt = self.src_dec_fmt = self.src_norm_fmt = ''
        self.outputs = WeakKeyDictionary[VideoOutput, vs.VideoNode]()
        self.tracking = False
        self._curr_frame_cache = WeakKeyDictionary[VideoOutput, tuple[int, vs.VideoFrame]]()
        self._curr_alphaframe_cache = WeakKeyDictionary[VideoOutput, tuple[int, vs.VideoFrame]]()
        self._mouse_is_subscribed = False

        self.last_pos: tuple[VideoOutput, QPoint] | None = None

        self.set_qobject_names()

        self.data_types = {
            vs.INTEGER: {
                1: ctypes.c_uint8,
                2: ctypes.c_uint16,
                4: ctypes.c_uint32,
            },
            vs.FLOAT: {
                2: ctypes.c_char,
                4: ctypes.c_float,
            }
        }

    def setup_ui(self) -> None:
        super().setup_ui()

        self.color_view = ColorView(self)
        self.color_view.setFixedSize(self.height() // 2, self.height() // 2)

        font = QFont('Consolas', 9)
        font.setStyleHint(QFont.StyleHint.Monospace)

        self.position = QLabel(self)

        self.rgb_label = QLabel('Rendered (RGB):', self)

        self.rgb_hex = QLabel(self)
        self.rgb_dec = QLabel(self)
        self.rgb_norm = QLabel(self)

        self.src_label = QLabel(self)

        self.src_hex = QLabel(self)
        self.src_dec = QLabel(self)
        self.src_norm = QLabel(self)

        self.rgb_hls = QLabel(self)
        self.rgb_hsv = QLabel(self)

        for label in [
            self.position,
            self.rgb_hex, self.rgb_dec, self.rgb_norm,
            self.src_hex, self.src_dec, self.src_norm
        ]:
            label.setFont(font)
            label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)

        self.copy_position_button = PushButton('⎘', self, clicked=self.on_copy_position_clicked)

        self.hlayout.addWidgets([
            QFrame(),
            self.color_view, self.position, self.copy_position_button,
            self.get_separator(),
            self.rgb_label, self.rgb_hex, self.rgb_dec, self.rgb_norm,
            self.get_separator(),
            self.src_label, self.src_hex, self.src_dec, self.src_norm,
            self.get_separator(),
            QLabel('HLS:'), self.rgb_hls, QLabel('HSV:'), self.rgb_hsv,
        ])

        self.hlayout.addStretch()

    def on_current_frame_changed(self, frame: Frame) -> None:
        if self.last_pos and self.last_pos[0] is self.main.current_output:
            self.update_labels(self.last_pos[1])

    def subscribe_on_mouse_events(self) -> None:
        if not self._mouse_is_subscribed:
            self.main.graphics_view.mouseMoved.connect(self.mouse_moved)
            self.main.graphics_view.mousePressed.connect(self.mouse_pressed)
            self.main.graphics_view.mouseReleased.connect(self.mouse_released)
        self._mouse_is_subscribed = True

    def unsubscribe_from_mouse_events(self) -> None:
        if self._mouse_is_subscribed:
            self.main.graphics_view.mouseMoved.disconnect(self.mouse_moved)
            self.main.graphics_view.mousePressed.disconnect(self.mouse_pressed)
            self.main.graphics_view.mouseReleased.disconnect(self.mouse_released)
        self._mouse_is_subscribed = False

    def mouse_moved(self, event: QMouseEvent) -> None:
        if self.tracking and not event.buttons():
            self.update_labels(event.pos())

    def mouse_pressed(self, event: QMouseEvent) -> None:
        if event.buttons() == Qt.MouseButton.RightButton:
            self.tracking = not self.tracking

        if self.tracking:
            self.update_labels(event.pos())

    def mouse_released(self, event: QMouseEvent) -> None:
        pass

    @property
    def current_source_frame(self) -> vs.VideoFrame:
        if self.main.current_output in self._curr_frame_cache:
            cache = self._curr_frame_cache[self.main.current_output]
        else:
            cache = None

        last_showed_frame = min(
            int(self.main.current_output.last_showed_frame), int(self.main.current_output.total_frames) - 1
        )

        if cache is None or cache[0] != last_showed_frame:
            cache = self._curr_frame_cache[self.main.current_output] = (
                last_showed_frame, self.outputs[self.main.current_output].get_frame(last_showed_frame)
            )

        return cache[1]

    @property
    def current_source_alpha_frame(self) -> vs.VideoFrame:
        assert self.main.current_output.source.alpha

        if self.main.current_output in self._curr_alphaframe_cache:
            cache = self._curr_alphaframe_cache[self.main.current_output]
        else:
            cache = None

        last_showed_frame = min(
            int(self.main.current_output.last_showed_frame), int(self.main.current_output.total_frames) - 1
        )

        if cache is None or cache[0] != last_showed_frame:
            cache = self._curr_frame_cache[self.main.current_output] = (
                last_showed_frame, self.main.current_output.source.alpha.get_frame(last_showed_frame)
            )

        return cache[1]

    def update_labels(self, local_pos: QPoint) -> None:
        self.last_pos = (self.main.current_output, local_pos)

        pos_f = self.main.graphics_view.mapToScene(local_pos)

        if not self.main.current_scene.contains(pos_f):
            return

        pos = QPoint(floor(pos_f.x()), floor(pos_f.y()))
        color = self.main.current_scene.pixmap().toImage().pixelColor(pos)
        components = color.red(), color.green(), color.blue()
        components_float = tuple[float, ...](x / 255 for x in components)

        self.color_view.color = color
        self.position.setText('{:4d},{:4d}'.format(pos.x(), pos.y()))

        self.rgb_hex.setText('{:2X},{:2X},{:2X}'.format(*components))
        self.rgb_dec.setText('{:3d},{:3d},{:3d}'.format(*components))
        self.rgb_norm.setText('{:0.5f},{:0.5f},{:0.5f}'.format(*components_float))
        self.rgb_hls.setText('{}%,{}%,{}%'.format(*(int(x * 255) for x in rgb_to_hls(*components_float))))
        self.rgb_hsv.setText('{}%,{}%,{}%'.format(*(int(x * 255) for x in rgb_to_hsv(*components_float))))

        if not self.src_label.isVisible():
            return

        fmt = self.current_source_frame.format

        src_vals = list(self.extract_value(self.current_source_frame, pos))
        if self.main.current_output.source.alpha:
            src_vals.append(next(self.extract_value(self.current_source_alpha_frame, pos)))

        self.src_dec.setText(self.src_dec_fmt.format(*src_vals))
        if fmt.sample_type == vs.INTEGER:
            self.src_hex.setText(self.src_hex_fmt.format(*src_vals))
            self.src_norm.setText(self.src_norm_fmt.format(*[
                src_val / self.src_max_val for src_val in src_vals
            ]))
        elif fmt.sample_type == vs.FLOAT:
            self.src_norm.setText(self.src_norm_fmt.format(*[
                max(0.0, min(val, 1.0)) if i in {0, 3} else max(-.5, min(val, .5)) + .5
                for i, val in enumerate(src_vals)
            ]))

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        super().on_current_output_changed(index, prev_index)

        if self.main.current_output not in self.outputs:
            self.outputs[self.main.current_output] = self.prepare_vs_output(self.main.current_output.source.clip)

        assert (src_fmt := self.outputs[self.main.current_output].format)

        has_alpha = bool(self.main.current_output.source.alpha)

        self.src_label.setText(f"Raw ({src_fmt.color_family.name}{' + Alpha' if has_alpha else ''}):")
        self.src_hex.setVisible(src_fmt.sample_type == vs.INTEGER)

        if src_fmt.sample_type == vs.INTEGER:
            self.src_max_val = 2**src_fmt.bits_per_sample - 1
        elif src_fmt.sample_type == vs.FLOAT:
            self.src_max_val = 1.0

        src_num_planes = src_fmt.num_planes + int(has_alpha)

        self.src_hex_fmt = ('{{:{w}X}},' * src_num_planes)[:-1].format(w=ceil(log(self.src_max_val, 16)))
        if src_fmt.sample_type == vs.INTEGER:
            self.src_dec_fmt = ('{{:{w}d}},' * src_num_planes)[:-1].format(w=ceil(log(self.src_max_val, 10)))
        elif src_fmt.sample_type == vs.FLOAT:
            self.src_dec_fmt = ('{: 0.5f},' * src_num_planes)[:-1]
        self.src_norm_fmt = ('{:0.5f},' * src_num_planes)[:-1]

        self.update_labels(self.main.graphics_view.mapFromGlobal(self.main.cursor().pos()))

    def on_copy_position_clicked(self, checked: bool | None = None) -> None:
        self.main.clipboard.setText(self.position.text().strip())

    def on_toggle(self, new_state: bool) -> None:
        super().on_toggle(new_state)
        self.tracking = new_state
        self.main.graphics_view.setMouseTracking(new_state)

        if new_state:
            self.subscribe_on_mouse_events()
            self.main.graphics_view.setDragMode(QGraphicsView.DragMode.NoDrag)
        else:
            self.unsubscribe_from_mouse_events()
            self.main.graphics_view.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)

    @staticmethod
    def prepare_vs_output(vs_output: vs.VideoNode) -> vs.VideoNode:
        assert (fmt := vs_output.format)

        if fmt.subsampling_w == fmt.subsampling_h == 0:
            return vs_output

        return vs.core.resize.Bicubic(
            vs_output, format=vs.core.query_video_format(
                fmt.color_family, fmt.sample_type, fmt.bits_per_sample, 0, 0
            ).id, dither_type='none'
        )

    def extract_value(self, vs_frame: vs.VideoFrame, pos: QPoint) -> Generator[float, None, None]:
        from ctypes import POINTER
        from ctypes import cast as ccast
        from struct import unpack

        fmt = vs_frame.format

        for plane in range(fmt.num_planes):
            stride = vs_frame.get_stride(plane)
            pointer = ccast(vs_frame.get_read_ptr(plane), POINTER(
                self.data_types[fmt.sample_type][fmt.bytes_per_sample] * (stride * vs_frame.height)  # type: ignore
            ))

            if fmt.sample_type == vs.FLOAT and fmt.bytes_per_sample == 2:
                offset = pos.y() * stride + pos.x() * fmt.bytes_per_sample
                yield cast(float, unpack('e', cast(bytearray, pointer.contents[
                    slice(offset, offset + fmt.bytes_per_sample)
                ]))[0])
            else:
                yield cast(int, pointer.contents[pos.y() * (stride // fmt.bytes_per_sample) + pos.x()])




from __future__ import annotations

from typing import Iterable, Iterator, Sequence, TypeVar, cast

from PyQt6.QtCore import QLineF, Qt
from PyQt6.QtGui import QColor
from vstools import fallback

from ..types import Frame, Scene, Time

__all__ = [
    'Notch',
    'Notches'
]


class Notch:
    def __init__(
        self, data: int | Frame | Time, color: QColor | Qt.GlobalColor | None = None,
        label: str | None = None, line: QLineF = QLineF()
    ) -> None:
        if isinstance(data, int):
            data = Frame(data)

        self.data = data
        self.color = cast(QColor, fallback(color, Qt.GlobalColor.white))
        self.label = fallback(label, '')
        self.line = line

    def __repr__(self) -> str:
        return '{}({}, {}, {}, {})'.format(
            type(self).__name__, repr(self.data), repr(self.color), repr(self.label), repr(self.line)
        )

    @classmethod
    def from_param(
        cls: type[NotchSelf], data: NotchT, color: QColor | Qt.GlobalColor | None = None, label: str | None = None
    ) -> Iterable[NotchSelf]:
        if isinstance(data, Notch):
            yield Notch(data.data, color if data.color is None else data.color, data.label or label, data.line)
            return

        if isinstance(data, Scene):
            if not label:
                label = data.label

            yield Notch(data.start, color, label)

            if data.end != data.start:
                yield Notch(data.end, color, label)

            return

        if isinstance(data, (int, Frame, Time)):
            yield Notch(data, color, label)
            return

        raise TypeError


NotchT = int | Frame | Scene | Time | Notch
NotchSelf = TypeVar('NotchSelf', bound=Notch)


class Notches:
    def __init__(
        self, other: Sequence[NotchT] | Notches | None = None,
        color: QColor | Qt.GlobalColor | None = None, label: str | None = None
    ) -> None:
        self.items = list[Notch]()

        if isinstance(other, Notches):
            self.items = list(other.items)
            return

        if isinstance(other, Sequence):
            for notch in other:
                self.add(notch, color, label)

    def add(
        self, data: NotchT, color: QColor | Qt.GlobalColor | None = None, label: str | None = None
    ) -> None:
        self.items.extend(Notch.from_param(data, color, label))

    def __len__(self) -> int:
        return len(self.items)

    def __getitem__(self, index: int) -> Notch:
        return self.items[index]

    def __iter__(self) -> Iterator[Notch]:
        return iter(self.items)

    def __repr__(self) -> str:
        return '{}({})'.format(type(self).__name__, repr(self.items))



from __future__ import annotations

import inspect

__all__ = [
    'is_preview'
]


def is_preview() -> bool:
    c_frame = [inspect.currentframe()]

    while (t := c_frame[-1] and c_frame[-1].f_back):
        c_frame.append(t)

        if t and '__name__' in t.f_locals:
            if t.f_locals['__name__'] == '__vspreview__':
                return True

    return False



from .vsenv import *  # noqa: F401, F403

from .abstracts import *  # noqa: F401, F403
from .bases import *  # noqa: F401, F403
from .custom import *  # noqa: F401, F403
from .types import *  # noqa: F401, F403



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QComboBox, QWidget

T = TypeVar('T')


__all__ = [
    'ComboBox'
]


class ComboBox(QComboBox, Generic[T]):
    def __class_getitem__(cls, _content_type: type[T]) -> type[ComboBox[T]]:
        class _ComboBox_inner(ComboBox):  # type: ignore
            content_type = _content_type

            valueChanged = pyqtSignal(content_type, object)

        return _ComboBox_inner

    indexChanged = pyqtSignal(int, int)
    content_type: type[T]

    if TYPE_CHECKING:
        valueChanged = pyqtSignal(content_type, T | None)  # type: ignore

    def __init__(self, parent: QWidget | None = None, **kwargs: Any) -> None:
        super().__init__(parent)

        self.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)

        self.oldValue = self.currentData()
        self.oldIndex = self.currentIndex()
        self.currentIndexChanged.connect(self._currentIndexChanged)

        for arg in kwargs:
            getattr(self, 'set' + arg[0].upper() + arg[1:])(kwargs.get(arg))

    def _currentIndexChanged(self, newIndex: int) -> None:
        if (newValue := self.currentData()) is None:
            return

        self.valueChanged.emit(newValue, self.oldValue)
        self.indexChanged.emit(newIndex, self.oldIndex)
        self.oldValue = newValue
        self.oldIndex = newIndex

    def currentValue(self) -> T:
        return cast(T, self.currentData())

    def setCurrentValue(self, newValue: T) -> None:
        self.setCurrentIndex(self.model().index_of(newValue))  # type: ignore



from __future__ import annotations

from enum import IntEnum, auto
from typing import TYPE_CHECKING, Any

from PyQt6.QtCore import QEvent, QPoint, QPointF, QRect, Qt, pyqtSignal
from PyQt6.QtGui import (
    QColor, QMouseEvent, QNativeGestureEvent, QPainter, QPalette, QPixmap, QResizeEvent, QTransform, QWheelEvent
)
from PyQt6.QtWidgets import (
    QApplication, QFrame, QGraphicsPixmapItem, QGraphicsScene, QGraphicsView, QSizePolicy, QWidget
)

if TYPE_CHECKING:
    from ...main import MainWindow
    from ..types import CroppingInfo


__all__ = [
    'DragEventType',
    'GraphicsView',
    'GraphicsImageItem',
    'MainVideoOutputGraphicsView'
]


class DragEventType(IntEnum):
    start = auto()
    stop = auto()
    move = auto()
    repaint = auto()


class GraphicsImageItem:
    __slots__ = ('_graphics_item', '_pixmap')

    def __init__(self, graphics_item: QGraphicsPixmapItem) -> None:
        self._graphics_item = graphics_item
        self._pixmap = self._graphics_item.pixmap()

    def contains(self, point: QPointF) -> bool:
        return self._graphics_item.contains(point)

    def hide(self) -> None:
        self._graphics_item.hide()

    def pixmap(self) -> QPixmap:
        return self._graphics_item.pixmap()

    def setPixmap(self, new_pixmap: QPixmap | None, crop_values: CroppingInfo | None = None) -> None:
        if new_pixmap is None:
            new_pixmap = self._pixmap
        else:
            self._pixmap = new_pixmap

        if crop_values is not None and crop_values.active:
            padded = QPixmap(new_pixmap.width(), new_pixmap.height())
            padded.fill(QColor(0, 0, 0, 0))
            painter = QPainter(padded)
            painter.drawPixmap(
                QPoint(crop_values.left, crop_values.top), new_pixmap,
                QRect(crop_values.left, crop_values.top, crop_values.width, crop_values.height)
            )
            painter.end()
            new_pixmap = padded

        self._graphics_item.setPixmap(new_pixmap)

    def show(self) -> None:
        self._graphics_item.show()


class GraphicsScene(QGraphicsScene):
    def __init__(self, view: GraphicsView) -> None:
        self.view = view
        self.main = self.view.main

        self.graphics_items = list[GraphicsImageItem]()

        super().__init__(self.main)

    def init_scenes(self) -> None:
        self.clear()
        self.graphics_items.clear()

        for _ in range(len(self.main.outputs)):
            raw_frame_item = self.addPixmap(QPixmap())
            raw_frame_item.hide()

            self.graphics_items.append(GraphicsImageItem(raw_frame_item))

    @property
    def current_scene(self) -> GraphicsImageItem:
        return self.graphics_items[self.main.current_output.index]


class GraphicsView(QGraphicsView):
    WHEEL_STEP = 15 * 8  # degrees

    __slots__ = ('app', 'angleRemainder', 'zoomValue',)

    mouseMoved = pyqtSignal(QMouseEvent)
    mousePressed = pyqtSignal(QMouseEvent)
    mouseReleased = pyqtSignal(QMouseEvent)
    wheelScrolled = pyqtSignal(int)
    dragEvent = pyqtSignal(DragEventType)
    drag_mode: QGraphicsView.DragMode

    underReload = False
    last_positions = (0, 0)

    _autofit = False
    main: MainWindow

    def __init__(self, main: MainWindow, parent: QWidget | None = None) -> None:
        from ...core import CheckBox, HBoxLayout
        from .combobox import ComboBox

        super().__init__(parent)

        self.main = main

        self.app = QApplication.instance()
        self.angleRemainder = 0
        self.zoomValue = 0.0
        self.currentZoom = 0.0
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        self.drag_mode = self.dragMode()

        self.setBackgroundBrush(self.main.palette().brush(QPalette.ColorRole.Window))
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)

        if self.main.settings.opengl_rendering_enabled:
            from PyQt6.QtOpenGLWidgets import QOpenGLWidget
            self.setViewport(QOpenGLWidget())

        self.wheelScrolled.connect(self.on_wheel_scrolled)
        self.main.reload_before_signal.connect(self.beforeReload)
        self.main.reload_after_signal.connect(self.afterReload)

        self.graphics_scene = GraphicsScene(self)
        self.setScene(self.graphics_scene)

        self.zoom_combobox = ComboBox[float](self, minimumContentsLength=4)

        self.auto_fit_button = CheckBox('Auto-fit', self, clicked=self.auto_fit_button_clicked)

        self.controls = QFrame()
        HBoxLayout(self.controls, [self.zoom_combobox, self.auto_fit_button])

        self.main.register_graphic_view(self)

        self.dragEvent.connect(self.propagate_move_event)

    def auto_fit_button_clicked(self, checked: bool) -> None:
        self.autofit = checked

    @property
    def current_scene(self) -> GraphicsImageItem:
        return self.graphics_scene.current_scene

    @property
    def autofit(self) -> bool:
        return self._autofit

    @autofit.setter
    def autofit(self, new_state: bool) -> None:
        self.zoom_combobox.setEnabled(not new_state)

        self._autofit = new_state

        if new_state:
            self.setZoom(None)
        else:
            self.setZoom(self.zoom_combobox.currentData())

    def bind_to(self, other_view: GraphicsView, *, mutual: bool = True) -> None:
        self.main.bound_graphics_views[other_view].add(self)

        if mutual:
            self.main.bound_graphics_views[self].add(other_view)

    def on_wheel_scrolled(self, steps: int) -> None:
        new_index = self.zoom_combobox.currentIndex() + steps

        if new_index < 0:
            new_index = 0
        elif new_index >= len(self.main.settings.zoom_levels):
            new_index = len(self.main.settings.zoom_levels) - 1

        self.zoom_combobox.setCurrentIndex(new_index)

    def setZoom(self, value: float | None) -> None:
        for view in self.main.bound_graphics_views[self]:
            view._setZoom(value)

    def _setZoom(self, value: float | None) -> None:
        if self.underReload or value == 0:
            return

        if value is None:
            if not self.autofit:
                return

            viewport = self.viewport()
            value = min(viewport.width() / self.content_width, viewport.height() / self.content_height)

        self.currentZoom = value / self.devicePixelRatio()

        self.setTransform(QTransform().scale(self.currentZoom, self.currentZoom))
        self.dragEvent.emit(DragEventType.repaint)

    def event(self, event: QEvent) -> bool:
        if self.underReload:
            event.ignore()
            return False

        if isinstance(event, QNativeGestureEvent):
            typ = event.gestureType()
            if typ == Qt.NativeGestureType.BeginNativeGesture:
                self.zoomValue = 0.0
            elif typ == Qt.NativeGestureType.ZoomNativeGesture:
                self.zoomValue += event.value()
            if typ == Qt.NativeGestureType.EndNativeGesture:
                self.wheelScrolled.emit(-1 if self.zoomValue < 0 else 1)

        return super().event(event)

    def wheelEvent(self, event: QWheelEvent) -> None:
        if self.underReload or self.autofit:
            return event.ignore()

        assert self.app

        modifier = event.modifiers()
        mouse = event.buttons()

        self.propagate_move_event()

        if modifier == Qt.KeyboardModifier.ControlModifier or mouse in {
            Qt.MouseButton.RightButton, Qt.MouseButton.MiddleButton
        }:
            angleDelta = event.angleDelta().y()

            # check if wheel wasn't rotated the other way since last rotation
            if self.angleRemainder * angleDelta < 0:
                self.angleRemainder = 0

            self.angleRemainder += angleDelta

            if abs(self.angleRemainder) >= self.WHEEL_STEP:
                self.wheelScrolled.emit(self.angleRemainder // self.WHEEL_STEP)
                self.angleRemainder %= self.WHEEL_STEP
            return
        elif modifier == Qt.KeyboardModifier.NoModifier:
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() - event.angleDelta().y())
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - event.angleDelta().x())
            return
        elif modifier == Qt.KeyboardModifier.ShiftModifier:
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() - event.angleDelta().x())
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - event.angleDelta().y())
            return

        event.ignore()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        if self.underReload or self.autofit:
            return event.ignore()

        super().mouseMoveEvent(event)

        if self.hasMouseTracking():
            self.mouseMoved.emit(event)

        self.dragEvent.emit(DragEventType.move)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if self.underReload or self.autofit:
            return event.ignore()

        if event.button() == Qt.MouseButton.LeftButton:
            self.drag_mode = self.dragMode()
            self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)

        super().mousePressEvent(event)

        self.mousePressed.emit(event)
        self.dragEvent.emit(DragEventType.start)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        if self.underReload or self.autofit:
            return event.ignore()

        super().mouseReleaseEvent(event)

        if event.button() == Qt.MouseButton.LeftButton:
            self.setDragMode(self.drag_mode)

        self.mouseReleased.emit(event)
        self.dragEvent.emit(DragEventType.stop)

    def beforeReload(self) -> None:
        self.underReload = True
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorViewCenter)
        self.last_positions = (self.verticalScrollBar().value(), self.horizontalScrollBar().value())

    def afterReload(self) -> None:
        self.underReload = False
        self.verticalScrollBar().setValue(self.last_positions[0])
        self.horizontalScrollBar().setValue(self.last_positions[1])
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setZoom(None)

    def propagate_move_event(self, _: Any = None) -> None:
        scrollbarW, scrollbarH = self.horizontalScrollBar(), self.verticalScrollBar()

        if not any({scrollbarW.isVisible(), scrollbarH.isVisible()}):
            return

        widthMax, heightMax = scrollbarW.maximum(), scrollbarH.maximum()

        for view in self.main.bound_graphics_views[self] - {self}:
            if view.isVisible():
                wBar, hBar = view.horizontalScrollBar(), view.verticalScrollBar()

                if widthMax:
                    wBar.setValue(int(scrollbarW.value() * wBar.maximum() / widthMax))

                if heightMax:
                    hBar.setValue(int(scrollbarH.value() * hBar.maximum() / heightMax))

    @property
    def content_width(self) -> int:
        raise NotImplementedError

    @property
    def content_height(self) -> int:
        raise NotImplementedError


class MainVideoOutputGraphicsView(GraphicsView):
    @property
    def content_width(self) -> int:
        return self.main.current_output.width

    @property
    def content_height(self) -> int:
        return self.main.current_output.height



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping

from PyQt6.QtCore import QKeyCombination, Qt
from PyQt6.QtWidgets import QComboBox, QFileDialog, QLabel, QSpacerItem

from ...core import (
    AbstractToolbar, CheckBox, CroppingInfo, HBoxLayout, LineEdit, PushButton, SpinBox, Stretch, Time, VBoxLayout,
    try_load
)
from ...core.custom import ComboBox, Switch
from ...models import GeneralModel
from ...utils import qt_silent_call
from .settings import MiscSettings

if TYPE_CHECKING:
    from ...main import MainWindow


__all__ = [
    'MiscToolbar'
]


class MiscToolbar(AbstractToolbar):
    __slots__ = (
        'reload_script_button',
        'save_storage_button', 'autosave_checkbox',
        'save_template_lineedit', 'save_frame_as_button',
        'toggle_button', 'save_file_types', 'copy_frame_button',
        'crop_top_spinbox', 'crop_left_spinbox', 'crop_width_spinbox',
        'crop_bottom_spinbox', 'crop_right_spinbox', 'crop_height_spinbox',
        'crop_active_switch', 'crop_mode_combox', 'crop_copycommand_button'
    )

    settings: MiscSettings

    def __init__(self, main: MainWindow) -> None:
        super().__init__(main, MiscSettings(self))

        self.setup_ui()

        self.save_file_types = {'Single Image (*.png)': self.save_as_png}

        self.main.settings.autosave_control.valueChanged.connect(self.on_autosave_interval_changed)

        self.add_shortcuts()

        self.set_qobject_names()

    def setup_ui(self) -> None:
        super().setup_ui()

        self.reload_script_button = PushButton('Reload Script', self, clicked=self.main.reload_script)

        self.save_storage_button = PushButton(
            'Save Storage', self, clicked=partial(self.main.dump_storage_async, manually=True)
        )

        self.autosave_checkbox = CheckBox('Autosave', self, checked=True)

        self.copy_frame_button = PushButton('Copy Frame', self, clicked=self.copy_frame_to_clipboard)

        self.save_frame_as_button = PushButton('Save Frame as', self, clicked=self.on_save_frame_as_clicked)

        self.save_template_lineedit = LineEdit(
            self.settings.SAVE_TEMPLATE, self, text=self.settings.SAVE_TEMPLATE,
            tooltip='''
                Available placeholders:
                    {format}, {fps_den}, {fps_num}, {frame},
                    {height}, {index}, {matrix}, {primaries}, {range},
                    {script_name}, {total_frames}, {transfer}, {width}.
                Frame props can be accessed as well using their names.
            '''.replace(' ' * 16, ' ').strip()
        )

        first_layer = [self.autosave_checkbox, self.get_separator()]

        if 'debug' in self.main.toolbars.toolbar_names:
            self.show_debug_checkbox = CheckBox('Show Debug Toolbar', self, stateChanged=self.on_show_debug_changed)
            first_layer.append(self.show_debug_checkbox)
            self.__slots__ = tuple([*self.__slots__, 'show_debug_checkbox'])  # type: ignore

        VBoxLayout(self.hlayout, [
            HBoxLayout([*first_layer, Stretch()]),
            HBoxLayout([
                self.reload_script_button, self.get_separator(),
                self.save_storage_button, self.get_separator(),
                self.copy_frame_button, Stretch()
            ]),
            HBoxLayout([self.save_frame_as_button, self.save_template_lineedit, Stretch()])
        ])

        self.hlayout.addStretch()
        self.hlayout.addStretch()

        self.crop_active_switch = Switch(10, 22, checked=True, clicked=self.crop_active_onchange)

        self.crop_top_spinbox = SpinBox(None, 0, 2 ** 16, valueChanged=self.crop_top_onchange)
        self.crop_left_spinbox = SpinBox(None, 0, 2 ** 16, valueChanged=self.crop_left_onchange)
        self.crop_bottom_spinbox = SpinBox(None, 0, 2 ** 16, valueChanged=self.crop_bottom_onchange)
        self.crop_right_spinbox = SpinBox(None, 0, 2 ** 16, valueChanged=self.crop_right_onchange)
        self.crop_width_spinbox = SpinBox(None, 1, 2 ** 16, valueChanged=self.crop_width_onchange)
        self.crop_height_spinbox = SpinBox(None, 1, 2 ** 16, valueChanged=self.crop_height_onchange)

        self.crop_copycommand_button = PushButton('Copy cropping command', clicked=self.crop_copycommand_onclick)

        self.crop_mode_combox = ComboBox[str](
            self, model=GeneralModel[str](['relative', 'absolute']),
            currentIndex=0, sizeAdjustPolicy=QComboBox.SizeAdjustPolicy.AdjustToContents
        )
        self.crop_mode_combox.currentIndexChanged.connect(self.crop_mode_onchange)

        self.crop_active_switch.click()

        HBoxLayout(self.hlayout, [
            VBoxLayout([
                HBoxLayout([
                    QLabel('Top'), self.crop_top_spinbox, QSpacerItem(35, 10)
                ], alignment=Qt.AlignmentFlag.AlignCenter, spacing=0),
                HBoxLayout([
                    QLabel('Left'), self.crop_left_spinbox,
                    self.crop_active_switch,
                    self.crop_right_spinbox, QLabel('Right')
                ], alignment=Qt.AlignmentFlag.AlignCenter, spacing=5),
                HBoxLayout([
                    QLabel('Bottom'), self.crop_bottom_spinbox, QSpacerItem(51, 10)
                ], alignment=Qt.AlignmentFlag.AlignCenter, spacing=0)
            ]),
            VBoxLayout([
                HBoxLayout([QLabel('Cropping Type:'), self.crop_mode_combox]),
                HBoxLayout([
                    QLabel('Width'), self.crop_width_spinbox,
                    QLabel('Height'), self.crop_height_spinbox
                ], spacing=0),
                HBoxLayout([self.crop_copycommand_button])
            ])
        ])

    def add_shortcuts(self) -> None:
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.ALT, Qt.Key.Key_S).toCombined(), self.save_storage_button.click
        )
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.CTRL, Qt.Key.Key_S).toCombined(), self.copy_frame_button.click
        )
        self.main.add_shortcut(
            QKeyCombination(Qt.Modifier.SHIFT, Qt.Key.Key_S).toCombined(), self.save_frame_as_button.click
        )

    def copy_frame_to_clipboard(self) -> None:
        self.main.clipboard.setPixmap(self.main.current_scene.pixmap())
        self.main.show_message('Current frame successfully copied to clipboard')

    def on_autosave_interval_changed(self, new_value: Time | None) -> None:
        if new_value is None:
            return
        if new_value == Time(seconds=0):
            self.main.autosave_timer.stop()
        else:
            self.main.autosave_timer.start(round(float(new_value) * 1000))

    def on_save_frame_as_clicked(self, checked: bool | None = None) -> None:
        from vstools import video_heuristics

        curr_out = self.main.current_output.source.clip
        fmt = curr_out.format
        assert fmt

        filter_str = ''.join([file_type + ';;' for file_type in self.save_file_types.keys()])[0:-2]

        template = self.save_template_lineedit.text()

        props = self.main.current_output.props

        heuristics = video_heuristics(self.main.current_output.source.clip, props)

        substitutions = {
            **props, **heuristics,
            'format': fmt.name,
            'fps_den': self.main.current_output.fps_den,
            'fps_num': self.main.current_output.fps_num,
            'width': self.main.current_output.width,
            'height': self.main.current_output.height,
            'script_name': self.main.script_path.stem,
            'index': self.main.current_output.index,
            'frame': self.main.current_output.last_showed_frame,
            'total_frames': self.main.current_output.total_frames
        }

        try:
            suggested_path_str = template.format(**substitutions)
        except KeyError:
            self.main.show_message('Save name template is invalid')
            return

        save_path_str, file_type = QFileDialog.getSaveFileName(
            self.main, 'Save as', suggested_path_str, filter_str
        )
        try:
            self.save_file_types[file_type](Path(save_path_str))
        except KeyError:
            pass

    def on_show_debug_changed(self, state: Qt.CheckState) -> None:
        assert hasattr(self.main.toolbars, 'debug')

        if state == Qt.CheckState.Checked:
            self.main.toolbars.debug.toggle_button.setVisible(True)
        elif state == Qt.CheckState.Unchecked:
            if self.main.toolbars.debug.toggle_button.isChecked():
                self.main.toolbars.debug.toggle_button.click()
            self.main.toolbars.debug.toggle_button.setVisible(False)

    def save_as_png(self, path: Path) -> None:
        self.main.current_scene.pixmap().save(str(path), 'PNG', self.main.settings.png_compression_level)

    def on_current_output_changed(self, index: int, prev_index: int) -> None:
        if index != prev_index:
            self.update_crop(prev_index)

        curr = self.main.current_output
        crop = curr.crop_values

        self.crop_top_spinbox.setMaximum(curr.height - 1)
        self.crop_bottom_spinbox.setMaximum(curr.height - 1)
        self.crop_left_spinbox.setMaximum(curr.width - 1)
        self.crop_right_spinbox.setMaximum(curr.width - 1)
        self.crop_width_spinbox.setMaximum(curr.width)
        self.crop_height_spinbox.setMaximum(curr.height)

        qt_silent_call(self.crop_top_spinbox.setValue, crop.top)
        qt_silent_call(self.crop_bottom_spinbox.setValue, curr.height - crop.height - crop.top)
        qt_silent_call(self.crop_left_spinbox.setValue, crop.left)
        qt_silent_call(self.crop_right_spinbox.setValue, curr.width - crop.width - crop.left)
        qt_silent_call(self.crop_width_spinbox.setValue, crop.width)
        qt_silent_call(self.crop_height_spinbox.setValue, crop.height)

        self.crop_active_switch.setChecked(not crop.active)
        self.crop_active_switch.click()

        self.crop_mode_combox.setCurrentIndex(int(crop.is_absolute))

    def update_crop(self, index: int | None = None) -> None:
        if not hasattr(self.main, 'current_output') or not self.main.outputs:
            return

        output = self.main.current_output if index is None else self.main.outputs[index]

        output.update_graphic_item(None, CroppingInfo(
            self.crop_top_spinbox.value(), self.crop_left_spinbox.value(),
            self.crop_width_spinbox.value(), self.crop_height_spinbox.value(),
            self.crop_active_switch.isChecked(), bool(self.crop_mode_combox.currentIndex())
        ), graphics_scene_item=self.main.current_output.graphics_scene_item)

    def crop_active_onchange(self, checked: bool) -> None:
        is_absolute = not not self.crop_mode_combox.currentIndex()

        self.crop_top_spinbox.setEnabled(checked)
        self.crop_left_spinbox.setEnabled(checked)

        self.crop_bottom_spinbox.setEnabled(checked and not is_absolute)
        self.crop_right_spinbox.setEnabled(checked and not is_absolute)

        self.crop_width_spinbox.setEnabled(checked and is_absolute)
        self.crop_height_spinbox.setEnabled(checked and is_absolute)

        self.crop_mode_combox.setEnabled(checked)

        self.crop_copycommand_button.setEnabled(checked)

        self.update_crop()

    def crop_mode_onchange(self, crop_mode_idx: int) -> None:
        is_absolute = bool(crop_mode_idx)

        self.crop_bottom_spinbox.setEnabled(not is_absolute)
        self.crop_right_spinbox.setEnabled(not is_absolute)

        self.crop_width_spinbox.setEnabled(is_absolute)
        self.crop_height_spinbox.setEnabled(is_absolute)

        self.update_crop()

    def crop_top_onchange(self, value: int) -> None:
        if not self.crop_active_switch.isChecked():
            return

        height = self.main.current_output.height
        offset = height - self.crop_bottom_spinbox.value()

        if offset - value < 1:
            qt_silent_call(self.crop_top_spinbox.setValue, offset - 1)
            return

        qt_silent_call(self.crop_height_spinbox.setValue, offset - value)

        self.update_crop()

    def crop_left_onchange(self, value: int) -> None:
        if not self.crop_active_switch.isChecked():
            return

        width = self.main.current_output.width
        offset = width - self.crop_right_spinbox.value()

        if offset - value < 1:
            qt_silent_call(self.crop_left_spinbox.setValue, offset - 1)
            return

        qt_silent_call(self.crop_width_spinbox.setValue, offset - value)

        self.update_crop()

    def crop_bottom_onchange(self, value: int) -> None:
        if not self.crop_active_switch.isChecked():
            return

        if self.crop_mode_combox.currentIndex():
            return

        height = self.main.current_output.height
        offset = height - self.crop_top_spinbox.value()

        if offset - value < 1:
            qt_silent_call(self.crop_bottom_spinbox.setValue, offset - 1)
            return

        qt_silent_call(self.crop_height_spinbox.setValue, offset - value)

        self.update_crop()

    def crop_right_onchange(self, value: int) -> None:
        if not self.crop_active_switch.isChecked():
            return

        if self.crop_mode_combox.currentIndex():
            return

        width = self.main.current_output.width
        offset = width - self.crop_left_spinbox.value()

        if offset - value < 1:
            qt_silent_call(self.crop_right_spinbox.setValue, offset - 1)
            return

        qt_silent_call(self.crop_width_spinbox.setValue, offset - value)

        self.update_crop()

    def crop_width_onchange(self, value: int) -> None:
        if not self.crop_active_switch.isChecked():
            return

        if self.crop_mode_combox.currentIndex():
            width = self.main.current_output.width
            offset = width - self.crop_left_spinbox.value()

            qt_silent_call(self.crop_right_spinbox.setValue, offset - value)

            if offset - value < 1:
                qt_silent_call(self.crop_left_spinbox.setValue, width - value)

        self.update_crop()

    def crop_height_onchange(self, value: int) -> None:
        if not self.crop_active_switch.isChecked():
            return

        if self.crop_mode_combox.currentIndex():
            height = self.main.current_output.height
            offset = height - self.crop_top_spinbox.value()

            qt_silent_call(self.crop_bottom_spinbox.setValue, offset - value)

            if offset - value < 1:
                qt_silent_call(self.crop_top_spinbox.setValue, height - value)

        self.update_crop()

    def crop_copycommand_onclick(self) -> None:
        is_absolute = self.crop_mode_combox.currentIndex()

        crop_top = self.crop_top_spinbox.value()
        crop_left = self.crop_left_spinbox.value()

        if not is_absolute:
            crop_bottom = self.crop_bottom_spinbox.value()
            crop_right = self.crop_right_spinbox.value()

            text = f'.std.Crop({crop_left}, {crop_right}, {crop_top}, {crop_bottom})'
        else:
            crop_width = self.crop_width_spinbox.value()
            crop_height = self.crop_height_spinbox.value()

            text = f'.std.CropAbs({crop_width}, {crop_height}, {crop_left}, {crop_top})'

        self.main.clipboard.setText(text)

    def __getstate__(self) -> Mapping[str, Any]:
        return super().__getstate__() | {
            'save_file_name_template': self.save_template_lineedit.text(),
            'show_debug': hasattr(self, 'show_debug_checkbox') and self.show_debug_checkbox.isChecked()
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try_load(state, 'save_file_name_template', str, self.save_template_lineedit.setText)

        if hasattr(self, 'show_debug_checkbox'):
            try_load(state, 'show_debug', bool, self.show_debug_checkbox.setChecked)

        super().__setstate__(state)



#!/usr/bin/env python3

import setuptools
from pathlib import Path

package_name = 'vspreview'

exec(Path(f'{package_name}/_metadata.py').read_text(), meta := dict[str, str]())

readme = Path('README.md').read_text()
requirements = Path('requirements.txt').read_text()


setuptools.setup(
    name=package_name,
    version=meta['__version__'],
    author=meta['__author_name__'],
    author_email=meta['__author_email__'],
    maintainer=meta['__maintainer_name__'],
    maintainer_email=meta['__maintainer_email__'],
    description=meta['__doc__'],
    long_description=readme,
    long_description_content_type='text/markdown',
    zip_safe=False,
    project_urls={
        'Source Code': 'https://github.com/Irrational-Encoding-Wizardry/vs-preview',
        'Documentation': 'https://vspreview.encode.moe/en/latest/',
        'Tracker': 'https://github.com/Irrational-Encoding-Wizardry/vs-preview/issues',
        'Contact': 'https://discord.gg/qxTxVJGtst'
    },
    install_requires=requirements,
    python_requires='>=3.10',
    packages=setuptools.find_packages('.', ('docs', 'stubs')),
    package_data={
        package_name: ['py.typed'],
        f'{package_name}.plugins': [
            file.name for file in (Path(package_name) / 'plugins').glob('*.ppy')
        ]
    },
    classifiers=[
        'Topic :: Multimedia :: Graphics',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'vspreview = vspreview.init:main'
        ]
    },
    command_options={
        "build_sphinx": {
            "project": ("setup.py", package_name),
            "version": ("setup.py", meta['__version__']),
            "release": ("setup.py", meta['__version__']),
            "source_dir": ("setup.py", "docs")
        }
    }
)



from .audio import *  # noqa: F401, F403
from .misc import *  # noqa: F401, F403
from .scene import *  # noqa: F401, F403
from .units import *  # noqa: F401, F403
from .video import *  # noqa: F401, F403
from .yaml import *  # noqa: F401, F403



from __future__ import annotations

from bisect import bisect_right
from copy import deepcopy
from typing import Any, Iterator, Mapping

from PyQt6.QtCore import QAbstractListModel, QAbstractTableModel, QModelIndex, Qt

from ..core import Frame, QYAMLObject, Scene, Time, main_window

__all__ = [
    'SceningList',
    'SceningLists'
]


class SceningList(QAbstractTableModel, QYAMLObject):
    __slots__ = ('name', 'items', 'max_value')

    START_FRAME_COLUMN = 0
    END_FRAME_COLUMN = 1
    START_TIME_COLUMN = 2
    END_TIME_COLUMN = 3
    LABEL_COLUMN = 4
    COLUMN_COUNT = 5

    def __init__(self, name: str = '', max_value: Frame | None = None, items: list[Scene] | None = None) -> None:
        self.setValue(name, max_value, items)

    def setValue(self, name: str = '', max_value: Frame | None = None, items: list[Scene] | None = None) -> None:
        super().__init__()
        self.name = name
        self.max_value = max_value if max_value is not None else Frame(2**31)
        self.items = items if items is not None else []

        self.main = main_window()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.items)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return self.COLUMN_COUNT

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        if orientation == Qt.Orientation.Horizontal:
            if section == self.START_FRAME_COLUMN:
                return 'Start'
            if section == self.END_FRAME_COLUMN:
                return 'End'
            if section == self.START_TIME_COLUMN:
                return 'Start'
            if section == self.END_TIME_COLUMN:
                return 'End'
            if section == self.LABEL_COLUMN:
                return 'Label'
        if orientation == Qt.Orientation.Vertical:
            return section + 1
        return None

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.UserRole) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        if row >= len(self.items):
            return None
        column = index.column()
        if column >= self.COLUMN_COUNT:
            return None

        if role in {Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole}:
            if column == self.START_FRAME_COLUMN:
                return str(self.items[row].start)
            if column == self.END_FRAME_COLUMN:
                if self.items[row].end != self.items[row].start:
                    return str(self.items[row].end)
                else:
                    return ''
            if column == self.START_TIME_COLUMN:
                return str(Time(self.items[row].start))
            if column == self.END_TIME_COLUMN:
                if self.items[row].end != self.items[row].start:
                    return str(Time(self.items[row].end))
                else:
                    return ''
            if column == self.LABEL_COLUMN:
                return str(self.items[row].label)

        if role == Qt.ItemDataRole.UserRole:
            if column == self.START_FRAME_COLUMN:
                return self.items[row].start
            if column == self.END_FRAME_COLUMN:
                return self.items[row].end
            if column == self.START_TIME_COLUMN:
                return Time(self.items[row].start)
            if column == self.END_TIME_COLUMN:
                return Time(self.items[row].end)
            if column == self.LABEL_COLUMN:
                return self.items[row].label

        return None

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if role not in {Qt.ItemDataRole.EditRole, Qt.ItemDataRole.UserRole}:
            return False

        row = index.row()
        column = index.column()
        scene = deepcopy(self.items[row])

        if column == self.START_FRAME_COLUMN:
            if not isinstance(value, Frame):
                raise TypeError
            if value > scene.end:
                return False
            scene.start = value
            proper_update = True
        elif column == self.END_FRAME_COLUMN:
            if not isinstance(value, Frame):
                raise TypeError
            if value < scene.start:
                return False
            scene.end = value
            proper_update = True
        if column == self.START_TIME_COLUMN:
            if not isinstance(value, Time):
                raise TypeError
            frame = Frame(value)
            if frame > scene.end:
                return False
            scene.start = frame
            proper_update = True
        if column == self.END_TIME_COLUMN:
            if not isinstance(value, Time):
                raise TypeError
            frame = Frame(value)
            if frame < scene.start:
                return False
            scene.end = frame
            proper_update = True
        elif column == self.LABEL_COLUMN:
            if not isinstance(value, str):
                raise TypeError
            scene.label = value
            proper_update = False

        if proper_update is True:
            i = bisect_right(self.items, scene)
            if i >= row:
                i -= 1
            if i != row:
                self.beginMoveRows(self.createIndex(row, 0), row, row, self.createIndex(i, 0), i)
                del self.items[row]
                self.items.insert(i, scene)
                self.endMoveRows()
            else:
                self.items[index.row()] = scene
                self.dataChanged.emit(index, index)
        else:
            self.items[index.row()] = scene
            self.dataChanged.emit(index, index)
        return True

    def __len__(self) -> int:
        return len(self.items)

    def __getitem__(self, i: int) -> Scene:
        return self.items[i]

    def __setitem__(self, i: int, value: Scene) -> None:
        if i >= len(self.items):
            raise IndexError

        self.items[i] = value
        self.dataChanged.emit(self.createIndex(i, 0), self.createIndex(i, self.COLUMN_COUNT - 1))

    def __contains__(self, item: Scene | Frame) -> bool:
        if isinstance(item, Scene):
            return item in self.items
        if isinstance(item, Frame):
            for scene in self.items:
                if item in scene:
                    return True
            return False
        raise TypeError

    def __iter__(self) -> Iterator[Scene]:
        return iter(self.items)

    def add(self, start: Frame, end: Frame | None = None, label: str = '') -> Scene:
        scene = Scene(start, end, label)

        if scene in self.items:
            return scene

        if scene.end > self.max_value:
            raise ValueError('New Scene is out of bounds of output')

        index = bisect_right(self.items, scene)
        self.beginInsertRows(QModelIndex(), index, index)
        self.items.insert(index, scene)
        self.endInsertRows()

        return scene

    def remove(self, i: int | Scene) -> None:
        if isinstance(i, Scene):
            i = self.items.index(i)

        if i >= 0 and i < len(self.items):
            self.beginRemoveRows(QModelIndex(), i, i)
            del self.items[i]
            self.endRemoveRows()
        else:
            raise IndexError

    def get_prev_frame(self, initial: Frame) -> Frame | None:
        result = None
        result_delta = Frame(int(self.max_value))
        for scene in reversed(self.items):
            if Frame(0) < initial - scene.start < result_delta:
                result = scene.start
                result_delta = scene.start - initial
            if Frame(0) < initial - scene.end < result_delta:
                result = scene.end
                result_delta = scene.end - initial

        return result

    def get_next_frame(self, initial: Frame) -> Frame | None:
        result = None
        result_delta = Frame(int(self.max_value))
        for scene in self.items:
            if Frame(0) < scene.start - initial < result_delta:
                result = scene.start
                result_delta = scene.start - initial
            if Frame(0) < scene.end - initial < result_delta:
                result = scene.end
                result_delta = scene.end - initial

        return result

    def __getstate__(self) -> Mapping[str, Any]:
        return {name: getattr(self, name)
                for name in self.__slots__}

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try:
            max_value = state['max_value']
            if not isinstance(max_value, Frame):
                raise TypeError("'max_value' of a SceningList is not a Frame. It's most probably corrupted.")

            name = state['name']
            if not isinstance(name, str):
                raise TypeError("'name' of a SceningList is not a Frame. It's most probably corrupted.")

            items = state['items']
            if not isinstance(items, list):
                raise TypeError("'items' of a SceningList is not a List. It's most probably corrupted.")
            for item in items:
                if not isinstance(item, Scene):
                    raise TypeError("One of the items of SceningList is not a Scene. It's most probably corrupted.")
        except KeyError:
            raise KeyError(
                "SceningList lacks one or more of its fields. It's most probably corrupted. Check those: {}.".format(
                    ', '.join(self.__slots__)))

        self.setValue(name, max_value, items)


class SceningLists(QAbstractListModel, QYAMLObject):
    __slots__ = ('items',)

    def __init__(self, items: list[SceningList] | None = None) -> None:
        self.setValue(items)

    def setValue(self, items: list[SceningList] | None = None) -> None:
        super().__init__()
        self.main = main_window()
        self.items = items if items is not None else []

    def __getitem__(self, i: int) -> SceningList:
        return self.items[i]

    def __len__(self) -> int:
        return len(self.items)

    def __iter__(self) -> Iterator[SceningList]:
        return iter(self.items)

    def index_of(self, item: SceningList, start_i: int = 0, end_i: int = 0) -> int:
        if end_i == 0:
            end_i = len(self.items)
        return self.items.index(item, start_i, end_i)

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self.items)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.UserRole) -> Any:
        if not index.isValid():
            return None
        if index.row() >= len(self.items):
            return None

        if role in {Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole}:
            return self.items[index.row()].name
        if role == Qt.ItemDataRole.UserRole:
            return self.items[index.row()]
        return None

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled

        return super().flags(index) | Qt.ItemFlag.ItemIsEditable

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:
        if not index.isValid():
            return False
        if role not in {Qt.ItemDataRole.EditRole, Qt.ItemDataRole.UserRole}:
            return False
        if not isinstance(value, str):
            return False

        self.items[index.row()].name = value
        self.dataChanged.emit(index, index)
        return True

    def insertRow(self, i: int, parent: QModelIndex = QModelIndex()) -> bool:
        self.add(i=i)
        return True

    def removeRow(self, i: int, parent: QModelIndex = QModelIndex()) -> bool:
        try:
            self.remove(i)
        except IndexError:
            return False

        return True

    def add(
        self, name: str | None = None, max_value: Frame | None = None, i: int | None = None
    ) -> tuple[SceningList, int]:
        if max_value is None:
            max_value = self.main.current_output.total_frames - 1
        if i is None:
            i = len(self.items)

        self.beginInsertRows(QModelIndex(), i, i)
        if name is None:
            self.items.insert(i, SceningList(f'List {len(self.items) + 1}', max_value))
        else:
            self.items.insert(i, SceningList(name, max_value))
        self.endInsertRows()
        return self.items[i], i

    def add_list(self, scening_list: SceningList) -> int:
        i = len(self.items)
        self.beginInsertRows(QModelIndex(), i, i)
        self.items.insert(i, scening_list)
        self.endInsertRows()
        return i

    def remove(self, item: int | SceningList) -> None:
        i = item
        if isinstance(i, SceningList):
            i = self.items.index(i)

        if i >= 0 and i < len(self.items):
            self.beginRemoveRows(QModelIndex(), i, i)
            del self.items[i]
            self.endRemoveRows()
        else:
            raise IndexError

    def __getstate__(self) -> Mapping[str, Any]:
        return {
            name: getattr(self, name)
            for name in self.__slots__
        }

    def __setstate__(self, state: Mapping[str, Any]) -> None:
        try:
            items = state['items']
            if not isinstance(items, list):
                raise TypeError('\'items\' of a SceningLists is not a List. It\'s most probably corrupted.')
            for item in items:
                if not isinstance(item, SceningList):
                    raise TypeError(
                        'One of the items of a SceningLists is not a SceningList. It\'s most probably corrupted.'
                    )
        except KeyError:
            raise KeyError(
                'SceningLists lacks one or more of its fields. It\'s most probably corrupted. Check those: {}.'
                .format(', '.join(self.__slots__))
            )

        self.setValue(items)
